@@ -1,22 +1,34 @@
 #!/usr/bin/perl
 
+use Getopt::Long;
 use strict;
 use warnings;
 use FileHandle;
 
 my $dumpchannels = 0;
 my $dumpdata = 0;
-while ($ARGV[0] =~ /^-/) {
-    my $opt = shift @ARGV;
-    if ($opt eq "--") {
-        last; # stop processing options
-    } elsif ($opt eq "-c") {
-        $dumpchannels = 1;
-    } elsif ($opt eq "-d") {
-        $dumpdata = 1;
-    } else {
-        die "unrecognised option '$opt'\n";
-    }
+my $pass_through_events = 0;
+my $verbose_all;
+my %verbose_packet;
+GetOptions("dump-channels|c" => \$dumpchannels,
+           "dump-data|d" => \$dumpdata,
+           "verbose|v" => \$verbose_all,
+           "full|f=s" => sub { $verbose_packet{$_[1]} = 1; },
+           "events|e" => \$pass_through_events,
+           "help" => sub { &usage(\*STDOUT, 0); })
+    or &usage(\*STDERR, 1);
+
+sub usage {
+    my ($fh, $exitstatus) = @_;
+    print $fh <<'EOF';
+usage:   logparse.pl [ options ] [ input-log-file ]
+options: --dump-channels, -c  dump the final state of every channel
+         --dump-data, -d      save data of every channel to ch0.i, ch0.o, ...
+         --full=PKT, -f PKT   print extra detail for packets of type PKT
+         --verbose, -v        print extra detail for all packets if available
+         --events, -e         copy Event Log messages from input log file
+EOF
+    exit $exitstatus;
 }
 
 my @channels = (); # ultimate channel ids are indices in this array
@@ -100,6 +112,41 @@ my %packets = (
         my ($direction, $seq, $data) = @_;
         print "\n";
     },
+#define SSH2_MSG_KEXGSS_INIT                      30  /* 0x1e */
+    'SSH2_MSG_KEXGSS_INIT' => sub {
+        my ($direction, $seq, $data) = @_;
+        print "\n";
+    },
+#define SSH2_MSG_KEXGSS_CONTINUE                  31  /* 0x1f */
+    'SSH2_MSG_KEXGSS_CONTINUE' => sub {
+        my ($direction, $seq, $data) = @_;
+        print "\n";
+    },
+#define SSH2_MSG_KEXGSS_COMPLETE                  32  /* 0x20 */
+    'SSH2_MSG_KEXGSS_COMPLETE' => sub {
+        my ($direction, $seq, $data) = @_;
+        print "\n";
+    },
+#define SSH2_MSG_KEXGSS_HOSTKEY                   33  /* 0x21 */
+    'SSH2_MSG_KEXGSS_HOSTKEY' => sub {
+        my ($direction, $seq, $data) = @_;
+        print "\n";
+    },
+#define SSH2_MSG_KEXGSS_ERROR                     34  /* 0x22 */
+    'SSH2_MSG_KEXGSS_ERROR' => sub {
+        my ($direction, $seq, $data) = @_;
+        print "\n";
+    },
+#define SSH2_MSG_KEXGSS_GROUPREQ                  40  /* 0x28 */
+    'SSH2_MSG_KEXGSS_GROUPREQ' => sub {
+        my ($direction, $seq, $data) = @_;
+        print "\n";
+    },
+#define SSH2_MSG_KEXGSS_GROUP                     41  /* 0x29 */
+    'SSH2_MSG_KEXGSS_GROUP' => sub {
+        my ($direction, $seq, $data) = @_;
+        print "\n";
+    },
 #define SSH2_MSG_KEXRSA_PUBKEY                    30    /* 0x1e */
     'SSH2_MSG_KEXRSA_PUBKEY' => sub {
         my ($direction, $seq, $data) = @_;
@@ -115,6 +162,16 @@ my %packets = (
         my ($direction, $seq, $data) = @_;
         print "\n";
     },
+#define SSH2_MSG_KEX_ECDH_INIT                    30    /* 0x1e */
+    'SSH2_MSG_KEX_ECDH_INIT' => sub {
+        my ($direction, $seq, $data) = @_;
+        print "\n";
+    },
+#define SSH2_MSG_KEX_ECDH_REPLY                   31    /* 0x1f */
+    'SSH2_MSG_KEX_ECDH_REPLY' => sub {
+        my ($direction, $seq, $data) = @_;
+        print "\n";
+    },
 #define SSH2_MSG_USERAUTH_REQUEST                 50	/* 0x32 */
     'SSH2_MSG_USERAUTH_REQUEST' => sub {
         my ($direction, $seq, $data) = @_;
@@ -523,6 +580,221 @@ my %packets = (
     },
 );
 
+our %disc_reasons = (
+    1 => "SSH_DISCONNECT_HOST_NOT_ALLOWED_TO_CONNECT",
+    2 => "SSH_DISCONNECT_PROTOCOL_ERROR",
+    3 => "SSH_DISCONNECT_KEY_EXCHANGE_FAILED",
+    4 => "SSH_DISCONNECT_RESERVED",
+    5 => "SSH_DISCONNECT_MAC_ERROR",
+    6 => "SSH_DISCONNECT_COMPRESSION_ERROR",
+    7 => "SSH_DISCONNECT_SERVICE_NOT_AVAILABLE",
+    8 => "SSH_DISCONNECT_PROTOCOL_VERSION_NOT_SUPPORTED",
+    9 => "SSH_DISCONNECT_HOST_KEY_NOT_VERIFIABLE",
+    10 => "SSH_DISCONNECT_CONNECTION_LOST",
+    11 => "SSH_DISCONNECT_BY_APPLICATION",
+    12 => "SSH_DISCONNECT_TOO_MANY_CONNECTIONS",
+    13 => "SSH_DISCONNECT_AUTH_CANCELLED_BY_USER",
+    14 => "SSH_DISCONNECT_NO_MORE_AUTH_METHODS_AVAILABLE",
+    15 => "SSH_DISCONNECT_ILLEGAL_USER_NAME",
+);
+
+my %verbose_packet_dump_functions = (
+    'SSH2_MSG_KEXINIT' => sub {
+        my ($data) = @_;
+        my ($cookie0, $cookie1, $cookie2, $cookie3,
+            $kex, $hostkey, $cscipher, $sccipher, $csmac, $scmac,
+            $cscompress, $sccompress, $cslang, $sclang, $guess, $reserved) =
+                &parse("uuuussssssssssbu", $data);
+        printf("  cookie: %08x%08x%08x%08x\n",
+               $cookie0, $cookie1, $cookie2, $cookie3);
+        my $print_namelist = sub {
+            my @names = split /,/, $_[1];
+            printf "  %s: name-list with %d items%s\n", $_[0], (scalar @names),
+            join "", map { "\n    $_" } @names;
+        };
+        $print_namelist->("kex", $kex);
+        $print_namelist->("host key", $hostkey);
+        $print_namelist->("client->server cipher", $cscipher);
+        $print_namelist->("server->client cipher", $sccipher);
+        $print_namelist->("client->server MAC", $csmac);
+        $print_namelist->("server->client MAC", $scmac);
+        $print_namelist->("client->server compression", $cscompress);
+        $print_namelist->("server->client compression", $sccompress);
+        $print_namelist->("client->server language", $cslang);
+        $print_namelist->("server->client language", $sclang);
+        printf "  first kex packet follows: %s\n", $guess;
+        printf "  reserved field: %#x\n", $reserved;
+    },
+    'SSH2_MSG_KEXDH_INIT' => sub {
+        my ($data) = @_;
+        my ($e) = &parse("m", $data);
+        printf "  e: %s\n", $e;
+    },
+    'SSH2_MSG_KEX_DH_GEX_REQUEST' => sub {
+        my ($data) = @_;
+        my ($min, $pref, $max) = &parse("uuu", $data);
+        printf "  min bits: %d\n", $min;
+        printf "  preferred bits: %d\n", $pref;
+        printf "  max bits: %d\n", $max;
+    },
+    'SSH2_MSG_KEX_DH_GEX_GROUP' => sub {
+        my ($data) = @_;
+        my ($p, $g) = &parse("mm", $data);
+        printf "  p: %s\n", $p;
+        printf "  g: %s\n", $g;
+    },
+    'SSH2_MSG_KEX_DH_GEX_INIT' => sub {
+        my ($data) = @_;
+        my ($e) = &parse("m", $data);
+        printf "  e: %s\n", $e;
+    },
+    'SSH2_MSG_KEX_ECDH_INIT' => sub {
+        my ($data) = @_;
+        my ($cpv) = &parse("s", $data);
+        # Public values in ECDH depend for their interpretation on the
+        # selected curve, and this script doesn't cross-analyse the
+        # two KEXINIT packets to independently figure out what that
+        # curve is. So the best we can do is just dump the raw data.
+        printf "  client public value: %s\n", (unpack "H*", $cpv);
+    },
+    'SSH2_MSG_KEXDH_REPLY' => sub {
+        my ($data) = @_;
+        my ($hostkeyblob, $f, $sigblob) = &parse("sms", $data);
+        my ($hktype, @hostkey) = &parse_public_key($hostkeyblob);
+        printf "  host key: %s\n", $hktype;
+        while (@hostkey) {
+            my ($key, $value) = splice @hostkey, 0, 2;
+            printf "    $key: $value\n";
+        }
+        printf "  f: %s\n", $f;
+        printf "  signature:\n";
+        my @signature = &parse_signature($sigblob, $hktype);
+        while (@signature) {
+            my ($key, $value) = splice @signature, 0, 2;
+            printf "    $key: $value\n";
+        }
+    },
+    'SSH2_MSG_KEX_DH_GEX_REPLY' => sub {
+        my ($data) = @_;
+        my ($hostkeyblob, $f, $sigblob) = &parse("sms", $data);
+        my ($hktype, @hostkey) = &parse_public_key($hostkeyblob);
+        printf "  host key: %s\n", $hktype;
+        while (@hostkey) {
+            my ($key, $value) = splice @hostkey, 0, 2;
+            printf "    $key: $value\n";
+        }
+        printf "  f: %s\n", $f;
+        printf "  signature:\n";
+        my @signature = &parse_signature($sigblob, $hktype);
+        while (@signature) {
+            my ($key, $value) = splice @signature, 0, 2;
+            printf "    $key: $value\n";
+        }
+    },
+    'SSH2_MSG_KEX_ECDH_REPLY' => sub {
+        my ($data) = @_;
+        my ($hostkeyblob, $spv, $sigblob) = &parse("sss", $data);
+        my ($hktype, @hostkey) = &parse_public_key($hostkeyblob);
+        printf "  host key: %s\n", $hktype;
+        while (@hostkey) {
+            my ($key, $value) = splice @hostkey, 0, 2;
+            printf "    $key: $value\n";
+        }
+        printf "  server public value: %s\n", (unpack "H*", $spv);
+        printf "  signature:\n";
+        my @signature = &parse_signature($sigblob, $hktype);
+        while (@signature) {
+            my ($key, $value) = splice @signature, 0, 2;
+            printf "    $key: $value\n";
+        }
+    },
+    'SSH2_MSG_NEWKEYS' => sub {},
+    'SSH2_MSG_SERVICE_REQUEST' => sub {
+        my ($data) = @_;
+        my ($servname) = &parse("s", $data);
+        printf "  service name: %s\n", $servname;
+    },
+    'SSH2_MSG_SERVICE_ACCEPT' => sub {
+        my ($data) = @_;
+        my ($servname) = &parse("s", $data);
+        printf "  service name: %s\n", $servname;
+    },
+    'SSH2_MSG_DISCONNECT' => sub {
+        my ($data) = @_;
+        my ($reason, $desc, $lang) = &parse("uss", $data);
+        printf("  reason code: %d%s\n", $reason,
+               defined $disc_reasons{$reason} ?
+               " ($disc_reasons{$reason})" : "" );
+        printf "  description: '%s'\n", $desc;
+        printf "  language tag: '%s'\n", $lang;
+    },
+    'SSH2_MSG_DEBUG' => sub {
+        my ($data) = @_;
+        my ($display, $desc, $lang) = &parse("bss", $data);
+        printf "  always display: %s\n", $display;
+        printf "  description: '%s'\n", $desc;
+        printf "  language tag: '%s'\n", $lang;
+    },
+    'SSH2_MSG_IGNORE' => sub {
+        my ($data) = @_;
+        my ($payload) = &parse("s", $data);
+        printf "  data: %s\n", unpack "H*", $payload;
+    },
+    'SSH2_MSG_UNIMPLEMENTED' => sub {
+        my ($data) = @_;
+        my ($seq) = &parse("u", $data);
+        printf "  sequence number: %d\n", $seq;
+    },
+    'SSH2_MSG_KEXGSS_INIT' => sub {
+        my ($data) = @_;
+        my ($token, $e) = &parse("sm", $data);
+        printf "  output token: %s\n", unpack "H*", $token;
+        printf "  e: %s\n", $e;
+    },
+    'SSH2_MSG_KEXGSS_CONTINUE' => sub {
+        my ($data) = @_;
+        my ($token) = &parse("s", $data);
+        printf "  output token: %s\n", unpack "H*", $token;
+    },
+    'SSH2_MSG_KEXGSS_COMPLETE' => sub {
+        my ($data) = @_;
+        my ($f, $permsgtoken, $got_output) = &parse("msb", $data);
+        printf "  f: %s\n", $f;
+        printf "  per-message token: %s\n", unpack "H*", $permsgtoken;
+        printf "  output token present: %s\n", $got_output;
+        if ($got_output eq "yes") {
+            my ($token) = &parse("s", $data);
+            printf "  output token: %s\n", unpack "H*", $token;
+        }
+    },
+    'SSH2_MSG_KEXGSS_HOSTKEY' => sub {
+        my ($data) = @_;
+        my ($hostkey) = &parse("s", $data);
+        printf "  host key: %s\n", unpack "H*", $hostkey;
+    },
+    'SSH2_MSG_KEXGSS_ERROR' => sub {
+        my ($data) = @_;
+        my ($maj, $min, $msg, $lang) = &parse("uuss", $data);
+        printf "  major status: %d\n", $maj;
+        printf "  minor status: %d\n", $min;
+        printf "  message: '%s'\n", $msg;
+        printf "  language tag: '%s'\n", $lang;
+    },
+    'SSH2_MSG_KEXGSS_GROUPREQ' => sub {
+        my ($data) = @_;
+        my ($min, $pref, $max) = &parse("uuu", $data);
+        printf "  min bits: %d\n", $min;
+        printf "  preferred bits: %d\n", $pref;
+        printf "  max bits: %d\n", $max;
+    },
+    'SSH2_MSG_KEXGSS_GROUP' => sub {
+        my ($data) = @_;
+        my ($p, $g) = &parse("mm", $data);
+        printf "  p: %s\n", $p;
+        printf "  g: %s\n", $g;
+    },
+);
+
 my %sftp_packets = (
 #define SSH_FXP_INIT                              1	/* 0x1 */
     0x1 => sub {
@@ -750,11 +1022,17 @@ my %sftp_packets = (
     },
 );
 
+for my $type (keys %verbose_packet) {
+    if (!defined $verbose_packet_dump_functions{$type}) {
+        die "no verbose dump available for packet type $type\n";
+    }
+}
+
 my ($direction, $seq, $ourseq, $type, $data, $recording);
 my %ourseqs = ('i'=>0, 'o'=>0);
 
 $recording = 0;
-while (<>) {
+while (<<>>) {
     if ($recording) {
         if (/^  [0-9a-fA-F]{8}  ((?:[0-9a-fA-F]{2} )*[0-9a-fA-F]{2})/) {
             push @$data, map { $_ eq "XX" ? -1 : hex $_ } split / /, $1;
@@ -762,11 +1040,26 @@ while (<>) {
             $recording = 0;
             my $fullseq = "$direction$ourseq";
             print "$fullseq: $type ";
+
+            my ($verbose_dump, $verbose_data) = undef;
+            if (defined $verbose_packet_dump_functions{$type} &&
+                ($verbose_all || defined $verbose_packet{$type})) {
+                $verbose_dump = $verbose_packet_dump_functions{$type};
+                $verbose_data = [ @$data ];
+            }
+
             if (defined $packets{$type}) {
                 $packets{$type}->($direction, $fullseq, $data);
             } else {
                 printf "raw %s\n", join "", map { sprintf "%02x", $_ } @$data;
             }
+            if (defined $verbose_dump) {
+                $verbose_dump->($verbose_data);
+                if (@$verbose_data) {
+                    printf("  trailing bytes: %s\n",
+                           unpack "H*", pack "C*", @$verbose_data);
+                }
+            }
         }
     }
     if (/^(Incoming|Outgoing) packet #0x([0-9a-fA-F]+), type \d+ \/ 0x[0-9a-fA-F]+ \((.*)\)/) {
@@ -781,6 +1074,9 @@ while (<>) {
         $data = [];
         $recording = 1;
     }
+    if ($pass_through_events && m/^Event Log: ([^\n]*)$/) {
+        printf "event: $1\n";
+    }
 }
 
 if ($dumpchannels) {
@@ -799,6 +1095,13 @@ if ($dumpchannels) {
     }
 }
 
+sub format_unsigned_hex_integer {
+    my $abs = join "", map { sprintf "%02x", $_ } @_;
+    $abs =~ s!^0*!!g;
+    $abs = "0" if $abs eq "";
+    return "0x" . $abs;
+}
+
 sub parseone {
     my ($type, $data) = @_;
     if ($type eq "u") { # uint32
@@ -824,7 +1127,7 @@ sub parseone {
         my $len = unpack "N", pack "C*", @bytes;
         @bytes = splice @$data, 0, $len;
         return "<missing>" if @bytes < $len or grep { $_<0 } @bytes;
-        if ($type eq "mpint") {
+        if ($type eq "m") {
             my $str = "";
             if ($bytes[0] >= 128) {
                 # Take two's complement.
@@ -839,7 +1142,7 @@ sub parseone {
                 }
                 $str = "-";
             }
-            $str .= "0x" . join "", map { sprintf "%02x", $_ } @bytes;
+            $str .= &format_unsigned_hex_integer(@bytes);
             return $str;
         } else {
             return pack "C*", @bytes;
@@ -944,6 +1247,80 @@ sub sftp_parse_attrs {
     return $out;
 }
 
+sub parse_public_key {
+    my ($blob) = @_;
+    my $data = [ unpack "C*", $blob ];
+    my @toret;
+    my ($type) = &parse("s", $data);
+    push @toret, $type;
+    if ($type eq "ssh-rsa") {
+        my ($e, $n) = &parse("mm", $data);
+        push @toret, "e", $e, "n", $n;
+    } elsif ($type eq "ssh-dss") {
+        my ($p, $q, $g, $y) = &parse("mmmm", $data);
+        push @toret, "p", $p, "q", $q, "g", $g, "y", $y;
+    } elsif ($type eq "ssh-ed25519") {
+        my ($xyblob) = &parse("s", $data);
+        my @y = unpack "C*", $xyblob;
+        push @toret, "hibit(x)", $y[$#y] & 1;
+        $y[$#y] &= ~1;
+        push @toret, "y & ~1", &format_unsigned_hex_integer(@y);
+    } elsif ($type =~ m!^ecdsa-sha2-nistp(256|384|521)$!) {
+        my ($curvename, $blob) = &parse("ss", $data);
+        push @toret, "curve name", $curvename;
+        my @blobdata = unpack "C*", $blob;
+        my ($fmt) = &parse("B", \@blobdata);
+        push @toret, "format byte", $fmt;
+        if ($fmt == 4) {
+            push @toret, "x", &format_unsigned_hex_integer(
+                @blobdata[0..($#blobdata+1)/2-1]);
+            push @toret, "y", &format_unsigned_hex_integer(
+                @blobdata[($#blobdata+1)/2..$#blobdata]);
+        }
+    } else {
+        push @toret, "undecoded data", unpack "H*", pack "C*", @$data;
+    }
+    return @toret;
+};
+
+sub parse_signature {
+    my ($blob, $keytype) = @_;
+    my $data = [ unpack "C*", $blob ];
+    my @toret;
+    if ($keytype eq "ssh-rsa") {
+        my ($type, $s) = &parse("ss", $data);
+        push @toret, "sig type", $type;
+        push @toret, "s", &format_unsigned_hex_integer(unpack "C*", $s);
+    } elsif ($keytype eq "ssh-dss") {
+        my ($type, $subblob) = &parse("ss", $data);
+        push @toret, "sig type", $type;
+        push @toret, "r", &format_unsigned_hex_integer(
+            unpack "C*", substr($subblob, 0, 20));
+        push @toret, "s", &format_unsigned_hex_integer(
+            unpack "C*", substr($subblob, 20, 40));
+    } elsif ($keytype eq "ssh-ed25519") {
+        my ($type, $rsblob) = &parse("ss", $data);
+        push @toret, "sig type", $type;
+        my @ry = unpack "C*", $rsblob;
+        my @sy = splice @ry, 32, 32;
+        push @toret, "hibit(r.x)", $ry[$#ry] & 1;
+        $ry[$#ry] &= ~1;
+        push @toret, "r.y & ~1", &format_unsigned_hex_integer(@ry);
+        push @toret, "hibit(s.x)", $sy[$#sy] & 1;
+        $sy[$#sy] &= ~1;
+        push @toret, "s.y & ~1", &format_unsigned_hex_integer(@sy);
+    } elsif ($keytype =~ m!^ecdsa-sha2-nistp(256|384|521)$!) {
+        my ($sigtype, $subblob) = &parse("ss", $data);
+        push @toret, "sig type", $sigtype;
+        my @sbdata = unpack "C*", $subblob;
+        my ($r, $s) = &parse("mm", \@sbdata);
+        push @toret, "r", $r, "s", $s;
+    } else {
+        push @toret, "undecoded data", unpack "H*", pack "C*", @$data;
+    }
+    return @toret;
+};
+
 sub stringescape {
     my ($str) = @_;
     $str =~ s!\\!\\\\!g;
