@@ -7,6 +7,7 @@
 
 #include <stdio.h>
 #include <ctype.h>
+#include <assert.h>
 
 #include "putty.h"
 #include "terminal.h"
@@ -14,19 +15,17 @@
 
 #define ECHOING (ldisc->localecho == FORCE_ON || \
                  (ldisc->localecho == AUTO && \
-                      (ldisc->back->ldisc(ldisc->backhandle, LD_ECHO) || \
-			   term_ldisc(ldisc->term, LD_ECHO))))
+                      (backend_ldisc_option_state(ldisc->backend, LD_ECHO))))
 #define EDITING (ldisc->localedit == FORCE_ON || \
                  (ldisc->localedit == AUTO && \
-                      (ldisc->back->ldisc(ldisc->backhandle, LD_EDIT) || \
-			   term_ldisc(ldisc->term, LD_EDIT))))
+                      (backend_ldisc_option_state(ldisc->backend, LD_EDIT))))
 
-static void c_write(Ldisc ldisc, char *buf, int len)
+static void c_write(Ldisc *ldisc, const void *buf, int len)
 {
-    from_backend(ldisc->frontend, 0, buf, len);
+    seat_stdout(ldisc->seat, buf, len);
 }
 
-static int plen(Ldisc ldisc, unsigned char c)
+static int plen(Ldisc *ldisc, unsigned char c)
 {
     if ((c >= 32 && c <= 126) || (c >= 160 && !in_utf(ldisc->term)))
 	return 1;
@@ -41,12 +40,12 @@ static int plen(Ldisc ldisc, unsigned char c)
 	return 4;		       /* <XY> hex representation */
 }
 
-static void pwrite(Ldisc ldisc, unsigned char c)
+static void pwrite(Ldisc *ldisc, unsigned char c)
 {
     if ((c >= 32 && c <= 126) ||
 	(!in_utf(ldisc->term) && c >= 0xA0) ||
 	(in_utf(ldisc->term) && c >= 0x80)) {
-	c_write(ldisc, (char *)&c, 1);
+	c_write(ldisc, &c, 1);
     } else if (c < 128) {
 	char cc[2];
 	cc[1] = (c == 127 ? '?' : c + 0x40);
@@ -59,15 +58,15 @@ static void pwrite(Ldisc ldisc, unsigned char c)
     }
 }
 
-static int char_start(Ldisc ldisc, unsigned char c)
+static bool char_start(Ldisc *ldisc, unsigned char c)
 {
     if (in_utf(ldisc->term))
 	return (c < 0x80 || c >= 0xC0);
     else
-	return 1;
+	return true;
 }
 
-static void bsb(Ldisc ldisc, int n)
+static void bsb(Ldisc *ldisc, int n)
 {
     while (n--)
 	c_write(ldisc, "\010 \010", 3);
@@ -76,77 +75,64 @@ static void bsb(Ldisc ldisc, int n)
 #define CTRL(x) (x^'@')
 #define KCTRL(x) ((x^'@') | 0x100)
 
-void *ldisc_create(Conf *conf, Terminal *term,
-		   Backend *back, void *backhandle,
-		   void *frontend)
+Ldisc *ldisc_create(Conf *conf, Terminal *term, Backend *backend, Seat *seat)
 {
-    Ldisc ldisc = snew(struct ldisc_tag);
+    Ldisc *ldisc = snew(Ldisc);
 
     ldisc->buf = NULL;
     ldisc->buflen = 0;
     ldisc->bufsiz = 0;
-    ldisc->quotenext = 0;
+    ldisc->quotenext = false;
 
-    ldisc->back = back;
-    ldisc->backhandle = backhandle;
+    ldisc->backend = backend;
     ldisc->term = term;
-    ldisc->frontend = frontend;
+    ldisc->seat = seat;
 
     ldisc_configure(ldisc, conf);
 
     /* Link ourselves into the backend and the terminal */
     if (term)
 	term->ldisc = ldisc;
-    if (back)
-	back->provide_ldisc(backhandle, ldisc);
+    if (backend)
+        backend_provide_ldisc(backend, ldisc);
 
     return ldisc;
 }
 
-void ldisc_configure(void *handle, Conf *conf)
+void ldisc_configure(Ldisc *ldisc, Conf *conf)
 {
-    Ldisc ldisc = (Ldisc) handle;
-
-    ldisc->telnet_keyboard = conf_get_int(conf, CONF_telnet_keyboard);
-    ldisc->telnet_newline = conf_get_int(conf, CONF_telnet_newline);
+    ldisc->telnet_keyboard = conf_get_bool(conf, CONF_telnet_keyboard);
+    ldisc->telnet_newline = conf_get_bool(conf, CONF_telnet_newline);
     ldisc->protocol = conf_get_int(conf, CONF_protocol);
     ldisc->localecho = conf_get_int(conf, CONF_localecho);
     ldisc->localedit = conf_get_int(conf, CONF_localedit);
 }
 
-void ldisc_free(void *handle)
+void ldisc_free(Ldisc *ldisc)
 {
-    Ldisc ldisc = (Ldisc) handle;
-
     if (ldisc->term)
 	ldisc->term->ldisc = NULL;
-    if (ldisc->back)
-	ldisc->back->provide_ldisc(ldisc->backhandle, NULL);
+    if (ldisc->backend)
+        backend_provide_ldisc(ldisc->backend, NULL);
     if (ldisc->buf)
 	sfree(ldisc->buf);
     sfree(ldisc);
 }
 
-void ldisc_send(void *handle, char *buf, int len, int interactive)
+void ldisc_echoedit_update(Ldisc *ldisc)
 {
-    Ldisc ldisc = (Ldisc) handle;
-    int keyflag = 0;
-    /*
-     * Called with len=0 when the options change. We must inform
-     * the front end in case it needs to know.
-     */
-    if (len == 0) {
-	ldisc_update(ldisc->frontend, ECHOING, EDITING);
-	return;
-    }
-    /*
-     * Notify the front end that something was pressed, in case
-     * it's depending on finding out (e.g. keypress termination for
-     * Close On Exit). 
-     */
-    frontend_keypress(ldisc->frontend);
+    seat_echoedit_update(ldisc->seat, ECHOING, EDITING);
+}
 
-    if (interactive && ldisc->term) {
+void ldisc_send(Ldisc *ldisc, const void *vbuf, int len, bool interactive)
+{
+    const char *buf = (const char *)vbuf;
+    int keyflag = 0;
+
+    assert(ldisc->term);
+    assert(len);
+
+    if (interactive) {
         /*
          * Interrupt a paste from the clipboard, if one was in
          * progress when the user pressed a key. This is easier than
@@ -221,7 +207,7 @@ void ldisc_send(void *handle, char *buf, int len, int interactive)
 			bsb(ldisc, plen(ldisc, ldisc->buf[ldisc->buflen - 1]));
 		    ldisc->buflen--;
 		}
-		ldisc->back->special(ldisc->backhandle, TS_EL);
+                backend_special(ldisc->backend, SS_EL, 0);
                 /*
                  * We don't send IP, SUSP or ABORT if the user has
                  * configured telnet specials off! This breaks
@@ -230,11 +216,11 @@ void ldisc_send(void *handle, char *buf, int len, int interactive)
                 if (!ldisc->telnet_keyboard)
                     goto default_case;
 		if (c == CTRL('C'))
-		    ldisc->back->special(ldisc->backhandle, TS_IP);
+                    backend_special(ldisc->backend, SS_IP, 0);
 		if (c == CTRL('Z'))
-		    ldisc->back->special(ldisc->backhandle, TS_SUSP);
+                    backend_special(ldisc->backend, SS_SUSP, 0);
 		if (c == CTRL('\\'))
-		    ldisc->back->special(ldisc->backhandle, TS_ABORT);
+                    backend_special(ldisc->backend, SS_ABORT, 0);
 		break;
 	      case CTRL('R'):	       /* redraw line */
 		if (ECHOING) {
@@ -245,13 +231,13 @@ void ldisc_send(void *handle, char *buf, int len, int interactive)
 		}
 		break;
 	      case CTRL('V'):	       /* quote next char */
-		ldisc->quotenext = TRUE;
+		ldisc->quotenext = true;
 		break;
 	      case CTRL('D'):	       /* logout or send */
 		if (ldisc->buflen == 0) {
-		    ldisc->back->special(ldisc->backhandle, TS_EOF);
+                    backend_special(ldisc->backend, SS_EOF, 0);
 		} else {
-		    ldisc->back->send(ldisc->backhandle, ldisc->buf, ldisc->buflen);
+                    backend_send(ldisc->backend, ldisc->buf, ldisc->buflen);
 		    ldisc->buflen = 0;
 		}
 		break;
@@ -287,13 +273,14 @@ void ldisc_send(void *handle, char *buf, int len, int interactive)
 		    /* FALLTHROUGH */
 	      case KCTRL('M'):	       /* send with newline */
 		    if (ldisc->buflen > 0)
-			ldisc->back->send(ldisc->backhandle, ldisc->buf, ldisc->buflen);
+                        backend_send(ldisc->backend,
+                                     ldisc->buf, ldisc->buflen);
 		    if (ldisc->protocol == PROT_RAW)
-			ldisc->back->send(ldisc->backhandle, "\r\n", 2);
+                        backend_send(ldisc->backend, "\r\n", 2);
 		    else if (ldisc->protocol == PROT_TELNET && ldisc->telnet_newline)
-			ldisc->back->special(ldisc->backhandle, TS_EOL);
+                        backend_special(ldisc->backend, SS_EOL, 0);
 		    else
-			ldisc->back->send(ldisc->backhandle, "\r", 1);
+                        backend_send(ldisc->backend, "\r", 1);
 		    if (ECHOING)
 			c_write(ldisc, "\r\n", 2);
 		    ldisc->buflen = 0;
@@ -302,20 +289,17 @@ void ldisc_send(void *handle, char *buf, int len, int interactive)
 		/* FALLTHROUGH */
 	      default:		       /* get to this label from ^V handler */
                 default_case:
-		if (ldisc->buflen >= ldisc->bufsiz) {
-		    ldisc->bufsiz = ldisc->buflen + 256;
-		    ldisc->buf = sresize(ldisc->buf, ldisc->bufsiz, char);
-		}
+                sgrowarray(ldisc->buf, ldisc->bufsiz, ldisc->buflen);
 		ldisc->buf[ldisc->buflen++] = c;
 		if (ECHOING)
 		    pwrite(ldisc, (unsigned char) c);
-		ldisc->quotenext = FALSE;
+		ldisc->quotenext = false;
 		break;
 	    }
 	}
     } else {
 	if (ldisc->buflen != 0) {
-	    ldisc->back->send(ldisc->backhandle, ldisc->buf, ldisc->buflen);
+            backend_send(ldisc->backend, ldisc->buf, ldisc->buflen);
 	    while (ldisc->buflen > 0) {
 		bsb(ldisc, plen(ldisc, ldisc->buf[ldisc->buflen - 1]));
 		ldisc->buflen--;
@@ -328,33 +312,33 @@ void ldisc_send(void *handle, char *buf, int len, int interactive)
 		switch (buf[0]) {
 		  case CTRL('M'):
 		    if (ldisc->protocol == PROT_TELNET && ldisc->telnet_newline)
-			ldisc->back->special(ldisc->backhandle, TS_EOL);
+                        backend_special(ldisc->backend, SS_EOL, 0);
 		    else
-			ldisc->back->send(ldisc->backhandle, "\r", 1);
+                        backend_send(ldisc->backend, "\r", 1);
 		    break;
 		  case CTRL('?'):
 		  case CTRL('H'):
 		    if (ldisc->telnet_keyboard) {
-			ldisc->back->special(ldisc->backhandle, TS_EC);
+                        backend_special(ldisc->backend, SS_EC, 0);
 			break;
 		    }
 		  case CTRL('C'):
 		    if (ldisc->telnet_keyboard) {
-			ldisc->back->special(ldisc->backhandle, TS_IP);
+                        backend_special(ldisc->backend, SS_IP, 0);
 			break;
 		    }
 		  case CTRL('Z'):
 		    if (ldisc->telnet_keyboard) {
-			ldisc->back->special(ldisc->backhandle, TS_SUSP);
+                        backend_special(ldisc->backend, SS_SUSP, 0);
 			break;
 		    }
 
 		  default:
-		    ldisc->back->send(ldisc->backhandle, buf, len);
+                    backend_send(ldisc->backend, buf, len);
 		    break;
 		}
 	    } else
-		ldisc->back->send(ldisc->backhandle, buf, len);
+                backend_send(ldisc->backend, buf, len);
 	}
     }
 }
