@@ -1,23 +1,15 @@
 /************************************************************************
- * $Id: minibidi.c 9426 2012-03-05 18:34:40Z simon $
  *
  * ------------
  * Description:
  * ------------
- * This is an implemention of Unicode's Bidirectional Algorithm
+ * This is an implementation of Unicode's Bidirectional Algorithm
  * (known as UAX #9).
  *
  *   http://www.unicode.org/reports/tr9/
  *
  * Author: Ahmad Khalifa
  *
- * -----------------
- * Revision Details:    (Updated by Revision Control System)
- * -----------------
- *  $Date: 2012-03-05 18:34:40 +0000 (Mon, 05 Mar 2012) $
- *  $Author: simon $
- *  $Revision: 9426 $
- *
  * (www.arabeyes.org - under MIT license)
  *
  ************************************************************************/
@@ -31,6 +23,7 @@
 
 #include <stdlib.h>	/* definition of wchar_t*/
 
+#include "putty.h"
 #include "misc.h"
 
 #define LMASK	0x3F	/* Embedding Level mask */
@@ -57,20 +50,16 @@ shapetypes[(xh)-SHAPE_FIRST].type : SU) /*))*/
 #define leastGreaterOdd(x) ( ((x)+1) | 1 )
 #define leastGreaterEven(x) ( ((x)+2) &~ 1 )
 
-typedef struct bidi_char {
-    unsigned int origwc, wc;
-    unsigned short index;
-} bidi_char;
-
 /* function declarations */
-void flipThisRun(bidi_char *from, unsigned char* level, int max, int count);
-int findIndexOfRun(unsigned char* level , int start, int count, int tlevel);
-unsigned char getType(int ch);
-unsigned char setOverrideBits(unsigned char level, unsigned char override);
-int getPreviousLevel(unsigned char* level, int from);
-int do_shape(bidi_char *line, bidi_char *to, int count);
-int do_bidi(bidi_char *line, int count);
-void doMirror(unsigned int *ch);
+static void flipThisRun(
+    bidi_char *from, unsigned char *level, int max, int count);
+static int findIndexOfRun(
+    unsigned char *level, int start, int count, int tlevel);
+static unsigned char getType(int ch);
+static unsigned char setOverrideBits(
+    unsigned char level, unsigned char override);
+static int getPreviousLevel(unsigned char *level, int from);
+static void doMirror(unsigned int *ch);
 
 /* character types */
 enum {
@@ -97,7 +86,7 @@ enum {
 
 /* Shaping Types */
 enum {
-    SL, /* Left-Joining, doesnt exist in U+0600 - U+06FF */
+    SL, /* Left-Joining, doesn't exist in U+0600 - U+06FF */
     SR, /* Right-Joining, ie has Isolated, Final */
     SD, /* Dual-Joining, ie has Isolated, Final, Initial, Medial */
     SU, /* Non-Joining */
@@ -305,7 +294,8 @@ const shape_node shapetypes[] = {
  * max: the maximum level found in this line (should be unsigned char)
  * count: line size in bidi_char
  */
-void flipThisRun(bidi_char *from, unsigned char *level, int max, int count)
+static void flipThisRun(
+    bidi_char *from, unsigned char *level, int max, int count)
 {
     int i, j, k, tlevel;
     bidi_char temp;
@@ -331,7 +321,8 @@ void flipThisRun(bidi_char *from, unsigned char *level, int max, int count)
 /*
  * Finds the index of a run with level equals tlevel
  */
-int findIndexOfRun(unsigned char* level , int start, int count, int tlevel)
+static int findIndexOfRun(
+    unsigned char *level , int start, int count, int tlevel)
 {
     int i;
     for (i=start; i<count; i++) {
@@ -363,7 +354,7 @@ perl -ne 'split ";"; $num = hex $_[0]; $type = $_[4];' \
     UnicodeData.txt
 
  */
-unsigned char getType(int ch)
+static unsigned char getType(int ch)
 {
     static const struct {
 	int first, last, type;
@@ -874,7 +865,8 @@ unsigned char getType(int ch)
         {0x4e00, 0x9fa5, L},
         {0xa000, 0xa48c, L},
         {0xac00, 0xd7a3, L},
-        {0xd800, 0xfa2d, L},
+        {0xd800, 0xdff7, L},
+        {0xe000, 0xfa2d, L},
         {0xfa30, 0xfa6a, L},
         {0xfb00, 0xfb06, L},
         {0xfb13, 0xfb17, L},
@@ -1032,7 +1024,7 @@ unsigned char getType(int ch)
  * important because all such characters are right-to-left so it
  * would have flagged them anyway.)
  */
-int is_rtl(int c)
+bool is_rtl(int c)
 {
     /*
      * After careful reading of the Unicode bidi algorithm (URL as
@@ -1053,7 +1045,8 @@ int is_rtl(int c)
  * This function sets the override bits of level according
  * to the value in override, and reurns the new byte.
  */
-unsigned char setOverrideBits(unsigned char level, unsigned char override)
+static unsigned char setOverrideBits(
+    unsigned char level, unsigned char override)
 {
     if (override == ON)
 	return level;
@@ -1069,7 +1062,7 @@ unsigned char setOverrideBits(unsigned char level, unsigned char override)
  * return the value _before_ it. Used to process U+202C POP
  * DIRECTIONAL FORMATTING.
  */
-int getPreviousLevel(unsigned char* level, int from)
+static int getPreviousLevel(unsigned char *level, int from)
 {
     if (from > 0) {
         unsigned char current = level[--from];
@@ -1094,9 +1087,10 @@ int getPreviousLevel(unsigned char* level, int from)
  */
 int do_shape(bidi_char *line, bidi_char *to, int count)
 {
-    int i, tempShape, ligFlag;
+    int i, tempShape;
+    bool ligFlag = false;
 
-    for (ligFlag=i=0; i<count; i++) {
+    for (i=0; i<count; i++) {
 	to[i] = line[i];
 	tempShape = STYPE(line[i].wc);
 	switch (tempShape) {
@@ -1121,28 +1115,28 @@ int do_shape(bidi_char *line, bidi_char *to, int count)
 	    if (line[i].wc == 0x644) {
 		if (i > 0) switch (line[i-1].wc) {
 		  case 0x622:
-		    ligFlag = 1;
+		    ligFlag = true;
 		    if ((tempShape == SL) || (tempShape == SD) || (tempShape == SC))
 			to[i].wc = 0xFEF6;
 		    else
 			to[i].wc = 0xFEF5;
 		    break;
 		  case 0x623:
-		    ligFlag = 1;
+		    ligFlag = true;
 		    if ((tempShape == SL) || (tempShape == SD) || (tempShape == SC))
 			to[i].wc = 0xFEF8;
 		    else
 			to[i].wc = 0xFEF7;
 		    break;
 		  case 0x625:
-		    ligFlag = 1;
+		    ligFlag = true;
 		    if ((tempShape == SL) || (tempShape == SD) || (tempShape == SC))
 			to[i].wc = 0xFEFA;
 		    else
 			to[i].wc = 0xFEF9;
 		    break;
 		  case 0x627:
-		    ligFlag = 1;
+		    ligFlag = true;
 		    if ((tempShape == SL) || (tempShape == SD) || (tempShape == SC))
 			to[i].wc = 0xFEFC;
 		    else
@@ -1151,7 +1145,7 @@ int do_shape(bidi_char *line, bidi_char *to, int count)
 		}
 		if (ligFlag) {
 		    to[i-1].wc = 0x20;
-		    ligFlag = 0;
+		    ligFlag = false;
 		    break;
 		}
 	    }
@@ -1194,18 +1188,19 @@ int do_bidi(bidi_char *line, int count)
     unsigned char currentEmbedding;
     unsigned char currentOverride;
     unsigned char tempType;
-    int i, j, yes, bover;
+    int i, j;
+    bool yes, bover;
 
     /* Check the presence of R or AL types as optimization */
-    yes = 0;
+    yes = false;
     for (i=0; i<count; i++) {
 	int type = getType(line[i].wc);
 	if (type == R || type == AL) {
-	    yes = 1;
+	    yes = true;
 	    break;
 	}
     }
-    if (yes == 0)
+    if (!yes)
 	return L;
 
     /* Initialize types, levels */
@@ -1258,7 +1253,7 @@ int do_bidi(bidi_char *line, int count)
      * terminated at the end of each paragraph. Paragraph separators are not
      * included in the embedding. (Useless here) NOT IMPLEMENTED
      */
-    bover = 0;
+    bover = false;
     for (i=0; i<count; i++) {
 	tempType = getType(line[i].wc);
 	switch (tempType) {
@@ -1277,13 +1272,13 @@ int do_bidi(bidi_char *line, int count)
 	  case RLO:
 	    currentEmbedding = levels[i] = leastGreaterOdd(currentEmbedding);
 	    tempType = currentOverride = R;
-	    bover = 1;
+	    bover = true;
 	    break;
 
 	  case LRO:
 	    currentEmbedding = levels[i] = leastGreaterEven(currentEmbedding);
 	    tempType = currentOverride = L;
-	    bover = 1;
+	    bover = true;
 	    break;
 
 	  case PDF:
@@ -1421,7 +1416,7 @@ int do_bidi(bidi_char *line, int count)
                 continue;
             } else if (i < count-1 && types[i+1] == ET) {
                 j=i;
-                while (j <count && types[j] == ET) {
+                while (j < count-1 && types[j] == ET) {
                     j++;
                 }
                 if (types[j] == EN)
@@ -1636,7 +1631,7 @@ int do_bidi(bidi_char *line, int count)
  * takes a pointer to a character that is checked for
  * having a mirror glyph.
  */
-void doMirror(unsigned int *ch)
+static void doMirror(unsigned int *ch)
 {
     if ((*ch & 0xFF00) == 0) {
 	switch (*ch) {
@@ -2022,7 +2017,7 @@ int main(int argc, char **argv)
 	unsigned long chr = strtoul(argv[i], NULL, 0);
 	int type = getType(chr);
 	assert(typetoname[type].type == type);
-	printf("U+%04x: %s\n", chr, typetoname[type].name);
+	printf("U+%04x: %s\n", (unsigned)chr, typetoname[type].name);
     }
 
     return 0;
