@@ -8,17 +8,16 @@
 #include <ctype.h>
 #include <string.h>
 
-#define DEFINE_PLUG_METHOD_MACROS
 #include "putty.h"
 #include "network.h"
 #include "proxy.h"
 
-void proxy_socks5_offerencryptedauth(char * command, int * len)
+void proxy_socks5_offerencryptedauth(BinarySink *bs)
 {
     /* For telnet, don't add any new encrypted authentication routines */
 }
 
-int proxy_socks5_handlechap (Proxy_Socket p)
+int proxy_socks5_handlechap (ProxySocket *p)
 {
 
     plug_closing(p->plug, "Proxy error: Trying to handle a SOCKS5 CHAP request"
@@ -27,7 +26,7 @@ int proxy_socks5_handlechap (Proxy_Socket p)
     return 1;
 }
 
-int proxy_socks5_selectchap(Proxy_Socket p)
+int proxy_socks5_selectchap(ProxySocket *p)
 {
     plug_closing(p->plug, "Proxy error: Trying to handle a SOCKS5 CHAP request"
 		 " in telnet-only build",
