@@ -1,37 +1,36 @@
 /*
- * pinger.c: centralised module that deals with sending TS_PING
+ * pinger.c: centralised module that deals with sending SS_PING
  * keepalives, to avoid replicating this code in multiple backends.
  */
 
 #include "putty.h"
 
-struct pinger_tag {
+struct Pinger {
     int interval;
-    int pending;
+    bool pending;
     unsigned long when_set, next;
-    Backend *back;
-    void *backhandle;
+    Backend *backend;
 };
 
-static void pinger_schedule(Pinger pinger);
+static void pinger_schedule(Pinger *pinger);
 
 static void pinger_timer(void *ctx, unsigned long now)
 {
-    Pinger pinger = (Pinger)ctx;
+    Pinger *pinger = (Pinger *)ctx;
 
     if (pinger->pending && now == pinger->next) {
-	pinger->back->special(pinger->backhandle, TS_PING);
-	pinger->pending = FALSE;
+        backend_special(pinger->backend, SS_PING, 0);
+	pinger->pending = false;
 	pinger_schedule(pinger);
     }
 }
 
-static void pinger_schedule(Pinger pinger)
+static void pinger_schedule(Pinger *pinger)
 {
     unsigned long next;
 
     if (!pinger->interval) {
-	pinger->pending = FALSE;       /* cancel any pending ping */
+	pinger->pending = false;       /* cancel any pending ping */
 	return;
     }
 
@@ -41,24 +40,23 @@ static void pinger_schedule(Pinger pinger)
         (next - pinger->when_set) < (pinger->next - pinger->when_set)) {
 	pinger->next = next;
         pinger->when_set = timing_last_clock();
-	pinger->pending = TRUE;
+	pinger->pending = true;
     }
 }
 
-Pinger pinger_new(Conf *conf, Backend *back, void *backhandle)
+Pinger *pinger_new(Conf *conf, Backend *backend)
 {
-    Pinger pinger = snew(struct pinger_tag);
+    Pinger *pinger = snew(Pinger);
 
     pinger->interval = conf_get_int(conf, CONF_ping_interval);
-    pinger->pending = FALSE;
-    pinger->back = back;
-    pinger->backhandle = backhandle;
+    pinger->pending = false;
+    pinger->backend = backend;
     pinger_schedule(pinger);
 
     return pinger;
 }
 
-void pinger_reconfig(Pinger pinger, Conf *oldconf, Conf *newconf)
+void pinger_reconfig(Pinger *pinger, Conf *oldconf, Conf *newconf)
 {
     int newinterval = conf_get_int(newconf, CONF_ping_interval);
     if (conf_get_int(oldconf, CONF_ping_interval) != newinterval) {
@@ -67,7 +65,7 @@ void pinger_reconfig(Pinger pinger, Conf *oldconf, Conf *newconf)
     }
 }
 
-void pinger_free(Pinger pinger)
+void pinger_free(Pinger *pinger)
 {
     expire_timer_context(pinger);
     sfree(pinger);
