@@ -15,11 +15,13 @@ my $setver = 0;
 my $upload = 0;
 my $precheck = 0;
 my $postcheck = 0;
+my $skip_ftp = 0;
 GetOptions("version=s" => \$version,
            "setver" => \$setver,
            "upload" => \$upload,
            "precheck" => \$precheck,
-           "postcheck" => \$postcheck)
+           "postcheck" => \$postcheck,
+           "no-ftp" => \$skip_ftp)
     or &usage();
 
 # --set-version: construct a local commit which updates the version
@@ -29,7 +31,6 @@ if ($setver) {
     0 == system "git", "diff-index", "--quiet", "--cached", "HEAD"
         or die "index is dirty";
     0 == system "git", "diff-files", "--quiet" or die "working tree is dirty";
-    -f "Makefile" and die "run 'make distclean' first";
     my $builddir = tempdir(DIR => ".", CLEANUP => 1);
     0 == system "git archive --format=tar HEAD | ( cd $builddir && tar xf - )"
         or die;
@@ -44,8 +45,6 @@ if ($setver) {
     $plink_transcript =~ s/^Unidentified build/Release ${version}/m or die;
     $plink_transcript =~ s/^/\\c /mg;
     &transform("LATEST.VER", sub { s/^\d+\.\d+$/$version/ });
-    &transform("windows/putty.iss", sub {
-        s/^(AppVerName=PuTTY version |VersionInfoTextVersion=Release |AppVersion=|VersionInfoVersion=)\d+\.\d+/$1$version/ });
     our $transforming = 0;
     &transform("doc/pscp.but", sub {
         if (/^\\c.*>pscp$/) { $transforming = 1; $_ .= $pscp_transcript; }
@@ -99,7 +98,7 @@ if ($upload) {
                        }
                    } elsif (m!^putty/(.*sum)s!) {
                        print $pipe "echo checking ${1}s\n";
-                       print $pipe "$1 -c ${1}s\n";
+                       print $pipe "grep -vF ' (installer version)' ${1}s | grep . | $1 -c\n";
                    }
                }, no_chdir => 1}, "putty");
         print $pipe "echo all verified ok\n";
@@ -164,11 +163,13 @@ if ($precheck || $postcheck) {
                    }
 
                    # Now test-download the files themselves.
-                   my $ftpdata = `curl -s $ftp_uri`;
-                   printf "  got %d bytes via FTP", length $ftpdata;
-                   die "FTP download for $ftp_uri did not match"
-                       if $ftpdata ne $real_content;
-                   print ", ok\n";
+                   unless ($skip_ftp) {
+                       my $ftpdata = `curl -s $ftp_uri`;
+                       printf "  got %d bytes via FTP", length $ftpdata;
+                       die "FTP download for $ftp_uri did not match"
+                           if $ftpdata ne $real_content;
+                       print ", ok\n";
+                   }
 
                    my $ua = LWP::UserAgent->new;
                    my $httpresponse = $ua->get($http_uri);
