@@ -16,19 +16,13 @@ struct beeptime {
     unsigned long ticks;
 };
 
+#define TRUST_SIGIL_WIDTH 3
+#define TRUST_SIGIL_CHAR 0xDFFE
+
 typedef struct {
     int y, x;
 } pos;
 
-#ifdef OPTIMISE_SCROLL
-struct scrollregion {
-    struct scrollregion *next;
-    int topline; /* Top line of scroll region. */
-    int botline; /* Bottom line of scroll region. */
-    int lines; /* Number of lines to scroll by - +ve is forwards. */
-};
-#endif /* OPTIMISE_SCROLL */
-
 typedef struct termchar termchar;
 typedef struct termline termline;
 
@@ -40,6 +34,7 @@ struct termchar {
      */
     unsigned long chr;
     unsigned long attr;
+    truecolour truecolour;
 
     /*
      * The cc_next field is used to link multiple termchars
@@ -60,17 +55,25 @@ struct termline {
     int cols;			       /* number of real columns on the line */
     int size;			       /* number of allocated termchars
 					* (cc-lists may make this > cols) */
-    int temporary;		       /* TRUE if decompressed from scrollback */
+    bool temporary;                    /* true if decompressed from scrollback */
     int cc_free;		       /* offset to first cc in free list */
     struct termchar *chars;
+    bool trusted;
 };
 
 struct bidi_cache_entry {
     int width;
+    bool trusted;
     struct termchar *chars;
     int *forward, *backward;	       /* the permutations of line positions */
 };
 
+struct term_utf8_decode {
+    int state;                         /* Is there a pending UTF-8 character */
+    int chr;                           /* and what is it so far? */
+    int size;                          /* The size of the UTF character. */
+};
+
 struct terminal_tag {
 
     int compatibility_level;
@@ -91,46 +94,40 @@ struct terminal_tag {
 
     struct beeptime *beephead, *beeptail;
     int nbeeps;
-    int beep_overloaded;
+    bool beep_overloaded;
     long lastbeep;
 
 #define TTYPE termchar
 #define TSIZE (sizeof(TTYPE))
 
-#ifdef OPTIMISE_SCROLL
-    struct scrollregion *scrollhead, *scrolltail;
-#endif /* OPTIMISE_SCROLL */
-
     int default_attr, curr_attr, save_attr;
+    truecolour curr_truecolour, save_truecolour;
     termchar basic_erase_char, erase_char;
 
     bufchain inbuf;		       /* terminal input buffer */
+
     pos curs;			       /* cursor */
     pos savecurs;		       /* saved cursor position */
     int marg_t, marg_b;		       /* scroll margins */
-    int dec_om;			       /* DEC origin mode flag */
-    int wrap, wrapnext;		       /* wrap flags */
-    int insert;			       /* insert-mode flag */
+    bool dec_om;                       /* DEC origin mode flag */
+    bool wrap, wrapnext;               /* wrap flags */
+    bool insert;                       /* insert-mode flag */
     int cset;			       /* 0 or 1: which char set */
     int save_cset, save_csattr;	       /* saved with cursor position */
-    int save_utf, save_wnext;	       /* saved with cursor position */
-    int rvideo;			       /* global reverse video flag */
+    bool save_utf, save_wnext;         /* saved with cursor position */
+    bool rvideo;                       /* global reverse video flag */
     unsigned long rvbell_startpoint;   /* for ESC[?5hESC[?5l vbell */
-    int cursor_on;		       /* cursor enabled flag */
-    int reset_132;		       /* Flag ESC c resets to 80 cols */
-    int use_bce;		       /* Use Background coloured erase */
-    int cblinker;		       /* When blinking is the cursor on ? */
-    int tblinker;		       /* When the blinking text is on */
-    int blink_is_real;		       /* Actually blink blinking text */
-    int term_echoing;		       /* Does terminal want local echo? */
-    int term_editing;		       /* Does terminal want local edit? */
+    bool cursor_on;                    /* cursor enabled flag */
+    bool reset_132;                    /* Flag ESC c resets to 80 cols */
+    bool use_bce;                      /* Use Background coloured erase */
+    bool cblinker;                     /* When blinking is the cursor on ? */
+    bool tblinker;                     /* When the blinking text is on */
+    bool blink_is_real;                /* Actually blink blinking text */
     int sco_acs, save_sco_acs;	       /* CSI 10,11,12m -> OEM charset */
-    int vt52_bold;		       /* Force bold on non-bold colours */
-    int utf;			       /* Are we in toggleable UTF-8 mode? */
-    int utf_state;		       /* Is there a pending UTF-8 character */
-    int utf_char;		       /* and what is it so far. */
-    int utf_size;		       /* The size of the UTF character. */
-    int printing, only_printing;       /* Are we doing ANSI printing? */
+    bool vt52_bold;                    /* Force bold on non-bold colours */
+    bool utf;                          /* Are we in toggleable UTF-8 mode? */
+    term_utf8_decode utf8;             /* If so, here's our decoding state */
+    bool printing, only_printing;      /* Are we doing ANSI printing? */
     int print_state;		       /* state of print-end-sequence scan */
     bufchain printer_buf;	       /* buffered data for printer */
     printer_job *print_job;
@@ -138,33 +135,38 @@ struct terminal_tag {
     /* ESC 7 saved state for the alternate screen */
     pos alt_savecurs;
     int alt_save_attr;
+    truecolour alt_save_truecolour;
     int alt_save_cset, alt_save_csattr;
-    int alt_save_utf, alt_save_wnext;
+    bool alt_save_utf;
+    bool alt_save_wnext;
     int alt_save_sco_acs;
 
     int rows, cols, savelines;
-    int has_focus;
-    int in_vbell;
+    bool has_focus;
+    bool in_vbell;
     long vbell_end;
-    int app_cursor_keys, app_keypad_keys, vt52_mode;
-    int repeat_off, cr_lf_return;
-    int seen_disp_event;
-    int big_cursor;
+    bool app_cursor_keys, app_keypad_keys, vt52_mode;
+    bool repeat_off, srm_echo, cr_lf_return;
+    bool seen_disp_event;
+    bool big_cursor;
 
     int xterm_mouse;		       /* send mouse messages to host */
-    int xterm_extended_mouse;
-    int urxvt_extended_mouse;
+    bool xterm_extended_mouse;
+    bool urxvt_extended_mouse;
     int mouse_is_down;		       /* used while tracking mouse buttons */
 
-    int bracketed_paste;
+    bool bracketed_paste, bracketed_paste_active;
 
     int cset_attr[2];
 
 /*
  * Saved settings on the alternate screen.
  */
-    int alt_x, alt_y, alt_om, alt_wrap, alt_wnext, alt_ins;
-    int alt_cset, alt_sco_acs, alt_utf;
+    int alt_x, alt_y;
+    bool alt_wnext, alt_ins;
+    bool alt_om, alt_wrap;
+    int alt_cset, alt_sco_acs;
+    bool alt_utf;
     int alt_t, alt_b;
     int alt_which;
     int alt_sblines; /* # of lines on alternate screen that should be used for scrollback. */
@@ -175,13 +177,13 @@ struct terminal_tag {
     unsigned esc_args[ARGS_MAX];
     int esc_nargs;
     int esc_query;
-#define ANSI(x,y)	((x)+((y)<<8))
-#define ANSI_QUE(x)	ANSI(x,TRUE)
+#define ANSI(x,y)	((x)+((y)*256))
+#define ANSI_QUE(x)	ANSI(x,1)
 
 #define OSC_STR_MAX 2048
     int osc_strlen;
     char osc_string[OSC_STR_MAX + 1];
-    int osc_w;
+    bool osc_w;
 
     char id_string[1024];
 
@@ -224,17 +226,18 @@ struct terminal_tag {
     wchar_t *paste_buffer;
     int paste_len, paste_pos;
 
-    void (*resize_fn)(void *, int, int);
-    void *resize_ctx;
+    Backend *backend;
 
-    void *ldisc;
+    Ldisc *ldisc;
 
-    void *frontend;
+    TermWin *win;
 
-    void *logctx;
+    LogContext *logctx;
 
     struct unicode_data *ucsdata;
 
+    unsigned long last_graphic_char;
+
     /*
      * We maintain a full copy of a Conf here, not merely a pointer
      * to it. That way, when we're passed a new one for
@@ -245,26 +248,26 @@ struct terminal_tag {
     Conf *conf;
 
     /*
-     * from_backend calls term_out, but it can also be called from
-     * the ldisc if the ldisc is called _within_ term_out. So we
-     * have to guard against re-entrancy - if from_backend is
-     * called recursively like this, it will simply add data to the
-     * end of the buffer term_out is in the process of working
-     * through.
+     * GUI implementations of seat_output call term_out, but it can
+     * also be called from the ldisc if the ldisc is called _within_
+     * term_out. So we have to guard against re-entrancy - if
+     * seat_output is called recursively like this, it will simply add
+     * data to the end of the buffer term_out is in the process of
+     * working through.
      */
-    int in_term_out;
+    bool in_term_out;
 
     /*
      * We schedule a window update shortly after receiving terminal
      * data. This tracks whether one is currently pending.
      */
-    int window_update_pending;
+    bool window_update_pending;
     long next_update;
 
     /*
      * Track pending blinks and tblinks.
      */
-    int tblink_pending, cblink_pending;
+    bool tblink_pending, cblink_pending;
     long next_tblink, next_cblink;
 
     /*
@@ -275,7 +278,13 @@ struct terminal_tag {
     bidi_char *wcFrom, *wcTo;
     int wcFromTo_size;
     struct bidi_cache_entry *pre_bidi_cache, *post_bidi_cache;
-    int bidi_cache_size;
+    size_t bidi_cache_size;
+
+    /*
+     * Current trust state, used to annotate every line of the
+     * terminal that a graphic character is output to.
+     */
+    bool trusted;
 
     /*
      * We copy a bunch of stuff out of the Conf structure into local
@@ -283,48 +292,97 @@ struct terminal_tag {
      * tree234 lookups which would be involved in fetching them from
      * the former every time.
      */
-    int ansi_colour;
+    bool ansi_colour;
     char *answerback;
     int answerbacklen;
-    int arabicshaping;
+    bool no_arabicshaping;
     int beep;
-    int bellovl;
+    bool bellovl;
     int bellovl_n;
     int bellovl_s;
     int bellovl_t;
-    int bidi;
-    int bksp_is_delete;
-    int blink_cur;
-    int blinktext;
-    int cjk_ambig_wide;
+    bool no_bidi;
+    bool bksp_is_delete;
+    bool blink_cur;
+    bool blinktext;
+    bool cjk_ambig_wide;
     int conf_height;
     int conf_width;
-    int crhaslf;
-    int erase_to_scrollback;
+    bool crhaslf;
+    bool erase_to_scrollback;
     int funky_type;
-    int lfhascr;
-    int logflush;
+    bool lfhascr;
+    bool logflush;
     int logtype;
-    int mouse_override;
-    int nethack_keypad;
-    int no_alt_screen;
-    int no_applic_c;
-    int no_applic_k;
-    int no_dbackspace;
-    int no_mouse_rep;
-    int no_remote_charset;
-    int no_remote_resize;
-    int no_remote_wintitle;
-    int no_remote_clearscroll;
-    int rawcnp;
-    int rect_select;
+    bool mouse_override;
+    bool nethack_keypad;
+    bool no_alt_screen;
+    bool no_applic_c;
+    bool no_applic_k;
+    bool no_dbackspace;
+    bool no_mouse_rep;
+    bool no_remote_charset;
+    bool no_remote_resize;
+    bool no_remote_wintitle;
+    bool no_remote_clearscroll;
+    bool rawcnp;
+    bool utf8linedraw;
+    bool rect_select;
     int remote_qtitle_action;
-    int rxvt_homeend;
-    int scroll_on_disp;
-    int scroll_on_key;
-    int xterm_256_colour;
+    bool rxvt_homeend;
+    bool scroll_on_disp;
+    bool scroll_on_key;
+    bool xterm_256_colour;
+    bool true_colour;
+
+    wchar_t *last_selected_text;
+    int *last_selected_attr;
+    truecolour *last_selected_tc;
+    size_t last_selected_len;
+    int mouse_select_clipboards[N_CLIPBOARDS];
+    int n_mouse_select_clipboards;
+    int mouse_paste_clipboard;
 };
 
-#define in_utf(term) ((term)->utf || (term)->ucsdata->line_codepage==CP_UTF8)
+static inline bool in_utf(Terminal *term)
+{
+    return term->utf || term->ucsdata->line_codepage == CP_UTF8;
+}
+
+unsigned long term_translate(
+    Terminal *term, term_utf8_decode *utf8, unsigned char c);
+static inline int term_char_width(Terminal *term, unsigned int c)
+{
+    return term->cjk_ambig_wide ? mk_wcwidth_cjk(c) : mk_wcwidth(c);
+}
+
+/*
+ * UCSINCOMPLETE is returned from term_translate if it's successfully
+ * absorbed a byte but not emitted a complete character yet.
+ * UCSTRUNCATED indicates a truncated multibyte sequence (so the
+ * caller emits an error character and then calls term_translate again
+ * with the same input byte). UCSINVALID indicates some other invalid
+ * multibyte sequence, such as an overlong synonym, or a standalone
+ * continuation byte, or a completely illegal thing like 0xFE. These
+ * values are not stored in the terminal data structures at all.
+ */
+#define UCSINCOMPLETE 0x8000003FU    /* '?' */
+#define UCSTRUNCATED  0x80000021U    /* '!' */
+#define UCSINVALID    0x8000002AU    /* '*' */
+
+/*
+ * Maximum number of combining characters we're willing to store in a
+ * character cell. Our linked-list data representation permits an
+ * unlimited number of these in principle, but if we allowed that in
+ * practice then it would be an easy DoS to just squirt a squillion
+ * identical combining characters to someone's terminal and cause
+ * their PuTTY or pterm to consume lots of memory and CPU pointlessly.
+ *
+ * The precise figure of 32 is more or less arbitrary, but one point
+ * supporting it is UAX #15's comment that 30 combining characters is
+ * "significantly beyond what is required for any linguistic or
+ * technical usage".
+ */
+#define CC_LIMIT 32
 
 #endif
