@@ -14,10 +14,13 @@
 #include <gdk/gdkkeysyms.h>
 #endif
 
+#include "defs.h"
 #include "gtkfont.h"
 #include "gtkcompat.h"
 #include "gtkmisc.h"
 
+#include "putty.h"
+#include "ssh.h"
 #include "misc.h"
 
 #define N_DRAWING_AREAS 3
@@ -27,7 +30,8 @@ struct drawing_area_ctx {
 #ifndef DRAW_DEFAULT_CAIRO
     GdkColor *cols;
 #endif
-    int width, height, current;
+    int width, height;
+    enum { NOT_CURRENT, CURRENT, GREYED_OUT } state;
 };
 
 struct askpass_ctx {
@@ -39,26 +43,63 @@ struct askpass_ctx {
 #endif
 #ifndef DRAW_DEFAULT_CAIRO
     GdkColormap *colmap;
-    GdkColor cols[2];
+    GdkColor cols[3];
 #endif
-    char *passphrase;
+    char *error_message;               /* if we finish without a passphrase */
+    char *passphrase;                  /* if we finish with one */
     int passlen, passsize;
 #if GTK_CHECK_VERSION(3,20,0)
     GdkSeat *seat;                     /* for gdk_seat_grab */
 #elif GTK_CHECK_VERSION(3,0,0)
     GdkDevice *keyboard;               /* for gdk_device_grab */
 #endif
+
+    int nattempts;
 };
 
+static prng *keypress_prng = NULL;
+static void feed_keypress_prng(void *data, int size)
+{
+    put_data(keypress_prng, data, size);
+}
+void random_add_noise(NoiseSourceId source, const void *noise, int length)
+{
+    if (keypress_prng)
+        prng_add_entropy(keypress_prng, source, make_ptrlen(noise, length));
+}
+static void setup_keypress_prng(void)
+{
+    keypress_prng = prng_new(&ssh_sha256);
+    prng_seed_begin(keypress_prng);
+    noise_get_heavy(feed_keypress_prng);
+    prng_seed_finish(keypress_prng);
+}
+static void cleanup_keypress_prng(void)
+{
+    prng_free(keypress_prng);
+}
+static int choose_new_area(int prev_area)
+{
+    /*
+     * Don't actually put the passphrase keystrokes themselves into
+     * the PRNG; that doesn't seem like the course of wisdom when
+     * that's precisely what the information displayed on the screen
+     * is trying _not_ to be correlated to.
+     */
+    noise_ultralight(NOISE_SOURCE_KEY, 0);
+    uint8_t data[8];
+    prng_read(keypress_prng, data, 8);
+    uint64_t randval = GET_64BIT_MSB_FIRST(data);
+    int reduced = randval % (N_DRAWING_AREAS - 1);
+    return (prev_area + 1 + reduced) % N_DRAWING_AREAS;
+}
+
 static void visually_acknowledge_keypress(struct askpass_ctx *ctx)
 {
-    int new_active;
-    new_active = rand() % (N_DRAWING_AREAS - 1);
-    if (new_active >= ctx->active_area)
-        new_active++;
-    ctx->drawingareas[ctx->active_area].current = 0;
+    int new_active = choose_new_area(ctx->active_area);
+    ctx->drawingareas[ctx->active_area].state = NOT_CURRENT;
     gtk_widget_queue_draw(ctx->drawingareas[ctx->active_area].area);
-    ctx->drawingareas[new_active].current = 1;
+    ctx->drawingareas[new_active].state = CURRENT;
     gtk_widget_queue_draw(ctx->drawingareas[new_active].area);
     ctx->active_area = new_active;
 }
@@ -113,11 +154,12 @@ static gint key_event(GtkWidget *widget, GdkEventKey *event, gpointer data)
                event->type == GDK_KEY_PRESS) {
         smemclr(ctx->passphrase, ctx->passsize);
         ctx->passphrase = NULL;
+        ctx->error_message = dupstr("passphrase input cancelled");
         gtk_main_quit();
     } else {
 #if GTK_CHECK_VERSION(2,0,0)
         if (gtk_im_context_filter_keypress(ctx->imc, event))
-            return TRUE;
+            return true;
 #endif
 
         if (event->type == GDK_KEY_PRESS) {
@@ -153,7 +195,7 @@ static gint key_event(GtkWidget *widget, GdkEventKey *event, gpointer data)
             }
         }
     }
-    return TRUE;
+    return true;
 }
 
 #if GTK_CHECK_VERSION(2,0,0)
@@ -172,21 +214,23 @@ static gint configure_area(GtkWidget *widget, GdkEventConfigure *event,
     ctx->width = event->width;
     ctx->height = event->height;
     gtk_widget_queue_draw(widget);
-    return TRUE;
+    return true;
 }
 
 #ifdef DRAW_DEFAULT_CAIRO
 static void askpass_redraw_cairo(cairo_t *cr, struct drawing_area_ctx *ctx)
 {
-    cairo_set_source_rgb(cr, 1-ctx->current, 1-ctx->current, 1-ctx->current);
+    double rgbval = (ctx->state == CURRENT ? 0 :
+                     ctx->state == NOT_CURRENT ? 1 : 0.5);
+    cairo_set_source_rgb(cr, rgbval, rgbval, rgbval);
     cairo_paint(cr);
 }
 #else
 static void askpass_redraw_gdk(GdkWindow *win, struct drawing_area_ctx *ctx)
 {
     GdkGC *gc = gdk_gc_new(win);
-    gdk_gc_set_foreground(gc, &ctx->cols[ctx->current]);
-    gdk_draw_rectangle(win, gc, TRUE, 0, 0, ctx->width, ctx->height);
+    gdk_gc_set_foreground(gc, &ctx->cols[ctx->state]);
+    gdk_draw_rectangle(win, gc, true, 0, 0, ctx->width, ctx->height);
     gdk_gc_unref(gc);
 }
 #endif
@@ -196,7 +240,7 @@ static gint draw_area(GtkWidget *widget, cairo_t *cr, gpointer data)
 {
     struct drawing_area_ctx *ctx = (struct drawing_area_ctx *)data;
     askpass_redraw_cairo(cr, ctx);
-    return TRUE;
+    return true;
 }
 #else
 static gint expose_area(GtkWidget *widget, GdkEventExpose *event,
@@ -212,13 +256,14 @@ static gint expose_area(GtkWidget *widget, GdkEventExpose *event,
     askpass_redraw_gdk(gtk_widget_get_window(ctx->area), ctx);
 #endif
 
-    return TRUE;
+    return true;
 }
 #endif
 
-static int try_grab_keyboard(struct askpass_ctx *ctx)
+static gboolean try_grab_keyboard(gpointer vctx)
 {
-    int ret;
+    struct askpass_ctx *ctx = (struct askpass_ctx *)vctx;
+    int i, ret;
 
 #if GTK_CHECK_VERSION(3,20,0)
     /*
@@ -226,16 +271,28 @@ static int try_grab_keyboard(struct askpass_ctx *ctx)
      * GdkSeat.
      */
     GdkSeat *seat;
+    GdkWindow *gdkw = gtk_widget_get_window(ctx->dialog);
+    if (!GDK_IS_WINDOW(gdkw) || !gdk_window_is_visible(gdkw))
+        goto fail;
 
     seat = gdk_display_get_default_seat
         (gtk_widget_get_display(ctx->dialog));
     if (!seat)
-        return FALSE;
+        goto fail;
 
     ctx->seat = seat;
-    ret = gdk_seat_grab(seat, gtk_widget_get_window(ctx->dialog),
-                        GDK_SEAT_CAPABILITY_KEYBOARD,
-                        TRUE, NULL, NULL, NULL, NULL);
+    ret = gdk_seat_grab(seat, gdkw, GDK_SEAT_CAPABILITY_KEYBOARD,
+                        true, NULL, NULL, NULL, NULL);
+
+    /*
+     * For some reason GDK 3.22 hides the GDK window as a side effect
+     * of a failed grab. I've no idea why. But if we're going to retry
+     * the grab, then we need to unhide it again or else we'll just
+     * get GDK_GRAB_NOT_VIEWABLE on every subsequent attempt.
+     */
+    if (ret != GDK_GRAB_SUCCESS)
+        gdk_window_show(gdkw);
+
 #elif GTK_CHECK_VERSION(3,0,0)
     /*
      * And it has to be done differently again prior to GTK 3.20.
@@ -246,22 +303,22 @@ static int try_grab_keyboard(struct askpass_ctx *ctx)
     dm = gdk_display_get_device_manager
         (gtk_widget_get_display(ctx->dialog));
     if (!dm)
-        return FALSE;
+        goto fail;
 
     pointer = gdk_device_manager_get_client_pointer(dm);
     if (!pointer)
-        return FALSE;
+        goto fail;
     keyboard = gdk_device_get_associated_device(pointer);
     if (!keyboard)
-        return FALSE;
+        goto fail;
     if (gdk_device_get_source(keyboard) != GDK_SOURCE_KEYBOARD)
-        return FALSE;
+        goto fail;
 
     ctx->keyboard = keyboard;
     ret = gdk_device_grab(ctx->keyboard,
                           gtk_widget_get_window(ctx->dialog),
                           GDK_OWNERSHIP_NONE,
-                          TRUE,
+                          true,
                           GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK,
                           NULL,
                           GDK_CURRENT_TIME);
@@ -270,37 +327,83 @@ static int try_grab_keyboard(struct askpass_ctx *ctx)
      * It's much simpler in GTK 1 and 2!
      */
     ret = gdk_keyboard_grab(gtk_widget_get_window(ctx->dialog),
-                            FALSE, GDK_CURRENT_TIME);
+                            false, GDK_CURRENT_TIME);
+#endif
+    if (ret != GDK_GRAB_SUCCESS)
+        goto fail;
+
+    /*
+     * Now that we've got the keyboard grab, connect up our keyboard
+     * handlers.
+     */
+#if GTK_CHECK_VERSION(2,0,0)
+    g_signal_connect(G_OBJECT(ctx->imc), "commit",
+                     G_CALLBACK(input_method_commit_event), ctx);
+#endif
+    g_signal_connect(G_OBJECT(ctx->dialog), "key_press_event",
+                     G_CALLBACK(key_event), ctx);
+    g_signal_connect(G_OBJECT(ctx->dialog), "key_release_event",
+                     G_CALLBACK(key_event), ctx);
+#if GTK_CHECK_VERSION(2,0,0)
+    gtk_im_context_set_client_window(ctx->imc,
+                                     gtk_widget_get_window(ctx->dialog));
 #endif
 
-    return ret == GDK_GRAB_SUCCESS;
+    /*
+     * And repaint the key-acknowledgment drawing areas as not greyed
+     * out.
+     */
+    ctx->active_area = rand() % N_DRAWING_AREAS;
+    for (i = 0; i < N_DRAWING_AREAS; i++) {
+        ctx->drawingareas[i].state =
+            (i == ctx->active_area ? CURRENT : NOT_CURRENT);
+        gtk_widget_queue_draw(ctx->drawingareas[i].area);
+    }
+
+    return false;
+
+  fail:
+    /*
+     * If we didn't get the grab, reschedule ourself on a timer to try
+     * again later.
+     *
+     * We have to do this rather than just trying once, because there
+     * is at least one important situation in which the grab may fail
+     * the first time: any user who is launching an add-key operation
+     * off some kind of window manager hotkey will almost by
+     * definition be running this script with a keyboard grab already
+     * active, namely the one-key grab that the WM (or whatever) uses
+     * to detect presses of the hotkey. So at the very least we have
+     * to give the user time to release that key.
+     */
+    if (++ctx->nattempts >= 4) {
+        smemclr(ctx->passphrase, ctx->passsize);
+        ctx->passphrase = NULL;
+        ctx->error_message = dupstr("unable to grab keyboard after 5 seconds");
+        gtk_main_quit();
+    } else {
+        g_timeout_add(1000/8, try_grab_keyboard, ctx);
+    }
+    return false;
 }
 
-typedef int (try_grab_fn_t)(struct askpass_ctx *ctx);
-
-static int repeatedly_try_grab(struct askpass_ctx *ctx, try_grab_fn_t fn)
+void realize(GtkWidget *widget, gpointer vctx)
 {
-    /*
-     * Repeatedly try to grab some aspect of the X server. We have to
-     * do this rather than just trying once, because there is at least
-     * one important situation in which the grab may fail the first
-     * time: any user who is launching an add-key operation off some
-     * kind of window manager hotkey will almost by definition be
-     * running this script with a keyboard grab already active, namely
-     * the one-key grab that the WM (or whatever) uses to detect
-     * presses of the hotkey. So at the very least we have to give the
-     * user time to release that key.
-     */
-    const useconds_t ms_limit = 5*1000000;  /* try for 5 seconds */
-    const useconds_t ms_step = 1000000/8;   /* at 1/8 second intervals */
-    useconds_t ms;
+    struct askpass_ctx *ctx = (struct askpass_ctx *)vctx;
 
-    for (ms = 0; ms < ms_limit; ms += ms_step) {
-        if (fn(ctx))
-            return TRUE;
-        usleep(ms_step);
-    }
-    return FALSE;
+    gtk_grab_add(ctx->dialog);
+
+    /*
+     * Schedule the first attempt at the keyboard grab.
+     */
+    ctx->nattempts = 0;
+#if GTK_CHECK_VERSION(3,20,0)
+    ctx->seat = NULL;
+#elif GTK_CHECK_VERSION(3,0,0)
+    ctx->keyboard = NULL;
+#endif
+
+    g_idle_add(try_grab_keyboard, ctx);
 }
 
 static const char *gtk_askpass_setup(struct askpass_ctx *ctx,
@@ -323,12 +426,12 @@ static const char *gtk_askpass_setup(struct askpass_ctx *ctx,
     ctx->promptlabel = gtk_label_new(prompt_text);
     align_label_left(GTK_LABEL(ctx->promptlabel));
     gtk_widget_show(ctx->promptlabel);
-    gtk_label_set_line_wrap(GTK_LABEL(ctx->promptlabel), TRUE);
+    gtk_label_set_line_wrap(GTK_LABEL(ctx->promptlabel), true);
 #if GTK_CHECK_VERSION(3,0,0)
     gtk_label_set_width_chars(GTK_LABEL(ctx->promptlabel), 48);
 #endif
     our_dialog_add_to_content_area(GTK_WINDOW(ctx->dialog),
-                                   ctx->promptlabel, TRUE, TRUE, 0);
+                                   ctx->promptlabel, true, true, 0);
 #if GTK_CHECK_VERSION(2,0,0)
     ctx->imc = gtk_im_multicontext_new();
 #endif
@@ -338,9 +441,10 @@ static const char *gtk_askpass_setup(struct askpass_ctx *ctx,
         ctx->colmap = gdk_colormap_get_system();
         ctx->cols[0].red = ctx->cols[0].green = ctx->cols[0].blue = 0xFFFF;
         ctx->cols[1].red = ctx->cols[1].green = ctx->cols[1].blue = 0;
+        ctx->cols[2].red = ctx->cols[2].green = ctx->cols[2].blue = 0x8000;
         gdk_colormap_alloc_colors(ctx->colmap, ctx->cols, 2,
-                                  FALSE, TRUE, success);
-        if (!success[0] | !success[1])
+                                  false, true, success);
+        if (!success[0] || !success[1])
             return "unable to allocate colours";
     }
 #endif
@@ -352,14 +456,14 @@ static const char *gtk_askpass_setup(struct askpass_ctx *ctx,
 #ifndef DRAW_DEFAULT_CAIRO
         ctx->drawingareas[i].cols = ctx->cols;
 #endif
-        ctx->drawingareas[i].current = 0;
+        ctx->drawingareas[i].state = GREYED_OUT;
         ctx->drawingareas[i].width = ctx->drawingareas[i].height = 0;
         /* It would be nice to choose this size in some more
          * context-sensitive way, like measuring the size of some
          * piece of template text. */
         gtk_widget_set_size_request(ctx->drawingareas[i].area, 32, 32);
         gtk_box_pack_end(action_area, ctx->drawingareas[i].area,
-                         TRUE, TRUE, 5);
+                         true, true, 5);
         g_signal_connect(G_OBJECT(ctx->drawingareas[i].area),
                          "configure_event",
                          G_CALLBACK(configure_area),
@@ -383,8 +487,7 @@ static const char *gtk_askpass_setup(struct askpass_ctx *ctx,
 
         gtk_widget_show(ctx->drawingareas[i].area);
     }
-    ctx->active_area = rand() % N_DRAWING_AREAS;
-    ctx->drawingareas[ctx->active_area].current = 1;
+    ctx->active_area = -1;
 
     /*
      * Arrange to receive key events. We don't really need to worry
@@ -393,40 +496,23 @@ static const char *gtk_askpass_setup(struct askpass_ctx *ctx,
      * the prompt label at random, and we'll use gtk_grab_add to
      * ensure key events go to it.
      */
-    gtk_widget_set_sensitive(ctx->promptlabel, TRUE);
+    gtk_widget_set_sensitive(ctx->dialog, true);
 
 #if GTK_CHECK_VERSION(2,0,0)
-    gtk_window_set_keep_above(GTK_WINDOW(ctx->dialog), TRUE);
+    gtk_window_set_keep_above(GTK_WINDOW(ctx->dialog), true);
 #endif
 
     /*
-     * Actually show the window, and wait for it to be shown.
+     * Wait for the key-receiving widget to actually be created, in
+     * order to call gtk_grab_add on it.
      */
-    gtk_widget_show_now(ctx->dialog);
+    g_signal_connect(G_OBJECT(ctx->dialog), "realize",
+                     G_CALLBACK(realize), ctx);
 
     /*
-     * Now that the window is displayed, make it grab the input focus.
+     * Show the window.
      */
-    gtk_grab_add(ctx->promptlabel);
-    if (!repeatedly_try_grab(ctx, try_grab_keyboard))
-        return "unable to grab keyboard";
-
-    /*
-     * And now that we've got the keyboard grab, connect up our
-     * keyboard handlers.
-     */
-#if GTK_CHECK_VERSION(2,0,0)
-    g_signal_connect(G_OBJECT(ctx->imc), "commit",
-                     G_CALLBACK(input_method_commit_event), ctx);
-#endif
-    g_signal_connect(G_OBJECT(ctx->promptlabel), "key_press_event",
-                     G_CALLBACK(key_event), ctx);
-    g_signal_connect(G_OBJECT(ctx->promptlabel), "key_release_event",
-                     G_CALLBACK(key_event), ctx);
-#if GTK_CHECK_VERSION(2,0,0)
-    gtk_im_context_set_client_window(ctx->imc,
-                                     gtk_widget_get_window(ctx->dialog));
-#endif
+    gtk_widget_show(ctx->dialog);
 
     return NULL;
 }
@@ -434,9 +520,11 @@ static const char *gtk_askpass_setup(struct askpass_ctx *ctx,
 static void gtk_askpass_cleanup(struct askpass_ctx *ctx)
 {
 #if GTK_CHECK_VERSION(3,20,0)
-    gdk_seat_ungrab(ctx->seat);
+    if (ctx->seat)
+        gdk_seat_ungrab(ctx->seat);
 #elif GTK_CHECK_VERSION(3,0,0)
-    gdk_device_ungrab(ctx->keyboard, GDK_CURRENT_TIME);
+    if (ctx->keyboard)
+        gdk_device_ungrab(ctx->keyboard, GDK_CURRENT_TIME);
 #else
     gdk_keyboard_ungrab(GDK_CURRENT_TIME);
 #endif
@@ -450,16 +538,16 @@ static void gtk_askpass_cleanup(struct askpass_ctx *ctx)
     gtk_widget_destroy(ctx->dialog);
 }
 
-static int setup_gtk(const char *display)
+static bool setup_gtk(const char *display)
 {
-    static int gtk_initialised = FALSE;
+    static bool gtk_initialised = false;
     int argc;
     char *real_argv[3];
     char **argv = real_argv;
-    int ret;
+    bool ret;
 
     if (gtk_initialised)
-        return TRUE;
+        return true;
 
     argc = 0;
     argv[argc++] = dupstr("dummy");
@@ -473,33 +561,38 @@ static int setup_gtk(const char *display)
     return ret;
 }
 
-const int buildinfo_gtk_relevant = TRUE;
+const bool buildinfo_gtk_relevant = true;
 
 char *gtk_askpass_main(const char *display, const char *wintitle,
-                       const char *prompt, int *success)
+                       const char *prompt, bool *success)
 {
     struct askpass_ctx actx, *ctx = &actx;
     const char *err;
 
+    ctx->passphrase = NULL;
+    ctx->error_message = NULL;
+
     /* In case gtk_init hasn't been called yet by the program */
     if (!setup_gtk(display)) {
-        *success = FALSE;
+        *success = false;
         return dupstr("unable to initialise GTK");
     }
 
     if ((err = gtk_askpass_setup(ctx, wintitle, prompt)) != NULL) {
-        *success = FALSE;
+        *success = false;
         return dupprintf("%s", err);
     }
+    setup_keypress_prng();
     gtk_main();
+    cleanup_keypress_prng();
     gtk_askpass_cleanup(ctx);
 
     if (ctx->passphrase) {
-        *success = TRUE;
+        *success = true;
         return ctx->passphrase;
     } else {
-        *success = FALSE;
-        return dupstr("passphrase input cancelled");
+        *success = false;
+        return ctx->error_message;
     }
 }
 
@@ -517,16 +610,16 @@ void modalfatalbox(const char *p, ...)
 
 int main(int argc, char **argv)
 {
-    int success, exitcode;
+    bool success;
+    int exitcode;
     char *ret;
 
     gtk_init(&argc, &argv);
 
     if (argc != 2) {
-        success = FALSE;
+        success = false;
         ret = dupprintf("usage: %s <prompt text>", argv[0]);
     } else {
-        srand(time(NULL));
         ret = gtk_askpass_main(NULL, "Enter passphrase", argv[1], &success);
     }
 
