@@ -13,26 +13,21 @@
 #include <unistd.h>
 #include <fcntl.h>
 #include <sys/time.h>
-#ifndef HAVE_NO_SYS_SELECT_H
-#include <sys/select.h>
-#endif
 
 #include "putty.h"
 #include "storage.h"
 #include "ssh.h"
 
-int console_batch_mode = FALSE;
-
-static void *console_logctx = NULL;
+bool console_batch_mode = false;
 
 static struct termios orig_termios_stderr;
-static int stderr_is_a_tty;
+static bool stderr_is_a_tty;
 
 void stderr_tty_init()
 {
     /* Ensure that if stderr is a tty, we can get it back to a sane state. */
     if ((flags & FLAG_STDERR_TTY) && isatty(STDERR_FILENO)) {
-	stderr_is_a_tty = TRUE;
+	stderr_is_a_tty = true;
 	tcgetattr(STDERR_FILENO, &orig_termios_stderr);
     }
 }
@@ -63,16 +58,58 @@ void cleanup_exit(int code)
     exit(code);
 }
 
-void set_busy_status(void *frontend, int status)
+/*
+ * Various error message and/or fatal exit functions.
+ */
+void console_print_error_msg(const char *prefix, const char *msg)
 {
+    struct termios cf;
+    premsg(&cf);
+    fputs(prefix, stderr);
+    fputs(": ", stderr);
+    fputs(msg, stderr);
+    fputc('\n', stderr);
+    fflush(stderr);
+    postmsg(&cf);
 }
 
-void update_specials_menu(void *frontend)
+void console_print_error_msg_fmt_v(
+    const char *prefix, const char *fmt, va_list ap)
 {
+    char *msg = dupvprintf(fmt, ap);
+    console_print_error_msg(prefix, msg);
+    sfree(msg);
 }
 
-void notify_remote_exit(void *frontend)
+void console_print_error_msg_fmt(const char *prefix, const char *fmt, ...)
 {
+    va_list ap;
+    va_start(ap, fmt);
+    console_print_error_msg_fmt_v(prefix, fmt, ap);
+    va_end(ap);
+}
+
+void modalfatalbox(const char *fmt, ...)
+{
+    va_list ap;
+    va_start(ap, fmt);
+    console_print_error_msg_fmt_v("FATAL ERROR", fmt, ap);
+    va_end(ap);
+    cleanup_exit(1);
+}
+
+void nonfatal(const char *fmt, ...)
+{
+    va_list ap;
+    va_start(ap, fmt);
+    console_print_error_msg_fmt_v("ERROR", fmt, ap);
+    va_end(ap);
+}
+
+void console_connection_fatal(Seat *seat, const char *msg)
+{
+    console_print_error_msg("FATAL ERROR", msg);
+    cleanup_exit(1);
 }
 
 void timer_change_notify(unsigned long next)
@@ -82,12 +119,13 @@ void timer_change_notify(unsigned long next)
 /*
  * Wrapper around Unix read(2), suitable for use on a file descriptor
  * that's been set into nonblocking mode. Handles EAGAIN/EWOULDBLOCK
- * by means of doing a one-fd select and then trying again; all other
- * errors (including errors from select) are returned to the caller.
+ * by means of doing a one-fd poll and then trying again; all other
+ * errors (including errors from poll) are returned to the caller.
  */
 static int block_and_read(int fd, void *buf, size_t len)
 {
     int ret;
+    pollwrapper *pw = pollwrap_new();
 
     while ((ret = read(fd, buf, len)) < 0 && (
 #ifdef EAGAIN
@@ -96,26 +134,29 @@ static int block_and_read(int fd, void *buf, size_t len)
 #ifdef EWOULDBLOCK
                (errno == EWOULDBLOCK) ||
 #endif
-               0)) {
+               false)) {
 
-        fd_set rfds;
-        FD_ZERO(&rfds);
-        FD_SET(fd, &rfds);
+        pollwrap_clear(pw);
+        pollwrap_add_fd_rwx(pw, fd, SELECT_R);
         do {
-            ret = select(fd+1, &rfds, NULL, NULL, NULL);
+            ret = pollwrap_poll_endless(pw);
         } while (ret < 0 && errno == EINTR);
         assert(ret != 0);
-        if (ret < 0)
+        if (ret < 0) {
+            pollwrap_free(pw);
             return ret;
-        assert(FD_ISSET(fd, &rfds));
+        }
+        assert(pollwrap_check_fd_rwx(pw, fd, SELECT_R));
     }
 
+    pollwrap_free(pw);
     return ret;
 }
 
-int verify_ssh_host_key(void *frontend, char *host, int port,
-                        const char *keytype, char *keystr, char *fingerprint,
-                        void (*callback)(void *ctx, int result), void *ctx)
+int console_verify_ssh_host_key(
+    Seat *seat, const char *host, int port,
+    const char *keytype, char *keystr, char *fingerprint,
+    void (*callback)(void *ctx, int result), void *ctx)
 {
     int ret;
 
@@ -219,12 +260,9 @@ int verify_ssh_host_key(void *frontend, char *host, int port,
     }
 }
 
-/*
- * Ask whether the selected algorithm is acceptable (since it was
- * below the configured 'warn' threshold).
- */
-int askalg(void *frontend, const char *algtype, const char *algname,
-	   void (*callback)(void *ctx, int result), void *ctx)
+int console_confirm_weak_crypto_primitive(
+    Seat *seat, const char *algtype, const char *algname,
+    void (*callback)(void *ctx, int result), void *ctx)
 {
     static const char msg[] =
 	"The first %s supported by the server is\n"
@@ -270,8 +308,9 @@ int askalg(void *frontend, const char *algtype, const char *algname,
     }
 }
 
-int askhk(void *frontend, const char *algname, const char *betteralgs,
-          void (*callback)(void *ctx, int result), void *ctx)
+int console_confirm_weak_cached_hostkey(
+    Seat *seat, const char *algname, const char *betteralgs,
+    void (*callback)(void *ctx, int result), void *ctx)
 {
     static const char msg[] =
 	"The first host key type we have stored for this server\n"
@@ -327,8 +366,9 @@ int askhk(void *frontend, const char *algname, const char *betteralgs,
  * Ask whether to wipe a session log file before writing to it.
  * Returns 2 for wipe, 1 for append, 0 for cancel (don't log).
  */
-int askappend(void *frontend, Filename *filename,
-	      void (*callback)(void *ctx, int result), void *ctx)
+static int console_askappend(LogPolicy *lp, Filename *filename,
+                             void (*callback)(void *ctx, int result),
+                             void *ctx)
 {
     static const char msgtemplate[] =
 	"The session log file \"%.*s\" already exists.\n"
@@ -376,6 +416,29 @@ int askappend(void *frontend, Filename *filename,
 	return 0;
 }
 
+bool console_antispoof_prompt = true;
+bool console_set_trust_status(Seat *seat, bool trusted)
+{
+    if (console_batch_mode || !is_interactive() || !console_antispoof_prompt) {
+        /*
+         * In batch mode, we don't need to worry about the server
+         * mimicking our interactive authentication, because the user
+         * already knows not to expect any.
+         *
+         * If standard input isn't connected to a terminal, likewise,
+         * because even if the server did send a spoof authentication
+         * prompt, the user couldn't respond to it via the terminal
+         * anyway.
+         *
+         * We also vacuously return success if the user has purposely
+         * disabled the antispoof prompt.
+         */
+        return true;
+    }
+
+    return false;
+}
+
 /*
  * Warn about the obsolescent key file format.
  * 
@@ -405,20 +468,30 @@ void old_keyfile_warning(void)
     postmsg(&cf);
 }
 
-void console_provide_logctx(void *logctx)
+static void console_logging_error(LogPolicy *lp, const char *string)
 {
-    console_logctx = logctx;
+    /* Errors setting up logging are considered important, so they're
+     * displayed to standard error even when not in verbose mode */
+    struct termios cf;
+    premsg(&cf);
+    fprintf(stderr, "%s\n", string);
+    fflush(stderr);
+    postmsg(&cf);
 }
 
-void logevent(void *frontend, const char *string)
+
+static void console_eventlog(LogPolicy *lp, const char *string)
 {
-    struct termios cf;
-    if ((flags & FLAG_STDERR) && (flags & FLAG_VERBOSE))
-        premsg(&cf);
-    if (console_logctx)
-	log_eventlog(console_logctx, string);
-    if ((flags & FLAG_STDERR) && (flags & FLAG_VERBOSE))
-        postmsg(&cf);
+    /* Ordinary Event Log entries are displayed in the same way as
+     * logging errors, but only in verbose mode */
+    if (flags & FLAG_VERBOSE)
+        console_logging_error(lp, string);
+}
+
+StripCtrlChars *console_stripctrl_new(
+    Seat *seat, BinarySink *bs_out, SeatInteractionContext sic)
+{
+    return stripctrl_new(bs_out, false, 0);
 }
 
 /*
@@ -446,18 +519,13 @@ static void console_close(FILE *outfp, int infd)
         fclose(outfp);             /* will automatically close infd too */
 }
 
-static void console_prompt_text(FILE *outfp, const char *data, int len)
+static void console_write(FILE *outfp, ptrlen data)
 {
-    int i;
-
-    for (i = 0; i < len; i++)
-	if ((data[i] & 0x60) || (data[i] == '\n'))
-	    fputc(data[i], outfp);
+    fwrite(data.ptr, 1, data.len, outfp);
     fflush(outfp);
 }
 
-int console_get_userpass_input(prompts_t *p, const unsigned char *in,
-                               int inlen)
+int console_get_userpass_input(prompts_t *p)
 {
     size_t curr_prompt;
     FILE *outfp = NULL;
@@ -482,17 +550,17 @@ int console_get_userpass_input(prompts_t *p, const unsigned char *in,
      */
     /* We only print the `name' caption if we have to... */
     if (p->name_reqd && p->name) {
-	size_t l = strlen(p->name);
-	console_prompt_text(outfp, p->name, l);
-	if (p->name[l-1] != '\n')
-	    console_prompt_text(outfp, "\n", 1);
+	ptrlen plname = ptrlen_from_asciz(p->name);
+	console_write(outfp, plname);
+        if (!ptrlen_endswith(plname, PTRLEN_LITERAL("\n"), NULL))
+	    console_write(outfp, PTRLEN_LITERAL("\n"));
     }
     /* ...but we always print any `instruction'. */
     if (p->instruction) {
-	size_t l = strlen(p->instruction);
-	console_prompt_text(outfp, p->instruction, l);
-	if (p->instruction[l-1] != '\n')
-	    console_prompt_text(outfp, "\n", 1);
+	ptrlen plinst = ptrlen_from_asciz(p->instruction);
+	console_write(outfp, plinst);
+        if (!ptrlen_endswith(plinst, PTRLEN_LITERAL("\n"), NULL))
+	    console_write(outfp, PTRLEN_LITERAL("\n"));
     }
 
     for (curr_prompt = 0; curr_prompt < p->n_prompts; curr_prompt++) {
@@ -510,7 +578,7 @@ int console_get_userpass_input(prompts_t *p, const unsigned char *in,
 	    newmode.c_lflag |= ECHO;
 	tcsetattr(infd, TCSANOW, &newmode);
 
-	console_prompt_text(outfp, pr->prompt, strlen(pr->prompt));
+	console_write(outfp, ptrlen_from_asciz(pr->prompt));
 
         len = 0;
         while (1) {
@@ -532,7 +600,7 @@ int console_get_userpass_input(prompts_t *p, const unsigned char *in,
 	tcsetattr(infd, TCSANOW, &oldmode);
 
 	if (!pr->echo)
-	    console_prompt_text(outfp, "\n", 1);
+            console_write(outfp, PTRLEN_LITERAL("\n"));
 
         if (len < 0) {
             console_close(outfp, infd);
@@ -547,15 +615,7 @@ int console_get_userpass_input(prompts_t *p, const unsigned char *in,
     return 1; /* success */
 }
 
-void frontend_keypress(void *handle)
-{
-    /*
-     * This is nothing but a stub, in console code.
-     */
-    return;
-}
-
-int is_interactive(void)
+bool is_interactive(void)
 {
     return isatty(0);
 }
@@ -567,3 +627,10 @@ int is_interactive(void)
 char *platform_get_x_display(void) {
     return dupstr(getenv("DISPLAY"));
 }
+
+static const LogPolicyVtable default_logpolicy_vt = {
+    console_eventlog,
+    console_askappend,
+    console_logging_error,
+};
+LogPolicy default_logpolicy[1] = {{ &default_logpolicy_vt }};
