@@ -16,8 +16,26 @@ char *get_random_data(int len, const char *device)
     int fd;
     int ngot, ret;
 
-    if (!device)
-        device = "/dev/random";
+    if (!device) {
+        static const char *const default_devices[] = {
+            "/dev/urandom", "/dev/random"
+        };
+        size_t i;
+
+        for (i = 0; i < lenof(default_devices); i++) {
+            if (access(default_devices[i], R_OK) == 0) {
+                device = default_devices[i];
+                break;
+            }
+        }
+
+        if (!device) {
+            sfree(buf);
+            fprintf(stderr, "puttygen: cannot find a readable "
+                    "random number source; use --random-device\n");
+            return NULL;
+        }
+    }
 
     fd = open(device, O_RDONLY);
     if (fd < 0) {
