@@ -41,11 +41,14 @@ static void gss_init(struct ssh_gss_library *lib, void *dlhandle,
     BIND_GSS_FN(delete_sec_context);
     BIND_GSS_FN(display_status);
     BIND_GSS_FN(get_mic);
+    BIND_GSS_FN(verify_mic);
     BIND_GSS_FN(import_name);
     BIND_GSS_FN(init_sec_context);
     BIND_GSS_FN(release_buffer);
     BIND_GSS_FN(release_cred);
     BIND_GSS_FN(release_name);
+    BIND_GSS_FN(acquire_cred);
+    BIND_GSS_FN(inquire_cred_by_mech);
 
 #undef BIND_GSS_FN
 
@@ -53,9 +56,10 @@ static void gss_init(struct ssh_gss_library *lib, void *dlhandle,
 }
 
 /* Dynamically load gssapi libs. */
-struct ssh_gss_liblist *ssh_gss_setup(const Config *cfg)
+struct ssh_gss_liblist *ssh_gss_setup(Conf *conf)
 {
     void *gsslib;
+    char *gsspath;
     struct ssh_gss_liblist *list = snew(struct ssh_gss_liblist);
 
     list->libraries = snewn(4, struct ssh_gss_library);
@@ -77,11 +81,11 @@ struct ssh_gss_liblist *ssh_gss_setup(const Config *cfg)
 		 2, "Using GSSAPI from libgss.so.1");
 
     /* User-specified GSSAPI library */
-    if (cfg->ssh_gss_custom.path[0] &&
-	(gsslib = dlopen(cfg->ssh_gss_custom.path, RTLD_LAZY)) != NULL)
+    gsspath = conf_get_filename(conf, CONF_ssh_gss_custom)->path;
+    if (*gsspath && (gsslib = dlopen(gsspath, RTLD_LAZY)) != NULL)
 	gss_init(&list->libraries[list->nlibraries++], gsslib,
 		 3, dupprintf("Using GSSAPI from user-specified"
-			      " library '%s'", cfg->ssh_gss_custom.path));
+			      " library '%s'", gsspath));
 
     return list;
 }
@@ -129,7 +133,7 @@ const struct keyvalwhere gsslibkeywords[] = {
 #include <gssapi/gssapi.h>
 
 /* Dynamically load gssapi libs. */
-struct ssh_gss_liblist *ssh_gss_setup(const Config *cfg)
+struct ssh_gss_liblist *ssh_gss_setup(Conf *conf)
 {
     struct ssh_gss_liblist *list = snew(struct ssh_gss_liblist);
 
@@ -144,11 +148,14 @@ struct ssh_gss_liblist *ssh_gss_setup(const Config *cfg)
     BIND_GSS_FN(delete_sec_context);
     BIND_GSS_FN(display_status);
     BIND_GSS_FN(get_mic);
+    BIND_GSS_FN(verify_mic);
     BIND_GSS_FN(import_name);
     BIND_GSS_FN(init_sec_context);
     BIND_GSS_FN(release_buffer);
     BIND_GSS_FN(release_cred);
     BIND_GSS_FN(release_name);
+    BIND_GSS_FN(acquire_cred);
+    BIND_GSS_FN(inquire_cred_by_mech);
 
 #undef BIND_GSS_FN
 
