@@ -14,9 +14,6 @@
 #include <pwd.h>
 #include <sys/ioctl.h>
 #include <sys/time.h>
-#ifndef HAVE_NO_SYS_SELECT_H
-#include <sys/select.h>
-#endif
 
 #define PUTTY_DO_GLOBALS	       /* actually _define_ globals */
 #include "putty.h"
@@ -25,92 +22,23 @@
 
 #define MAX_STDIN_BACKLOG 4096
 
-static void *logctx;
+static LogContext *logctx;
 
 static struct termios orig_termios;
 
-void fatalbox(const char *p, ...)
+void cmdline_error(const char *fmt, ...)
 {
-    struct termios cf;
     va_list ap;
-    premsg(&cf);
-    fprintf(stderr, "FATAL ERROR: ");
-    va_start(ap, p);
-    vfprintf(stderr, p, ap);
+    va_start(ap, fmt);
+    console_print_error_msg_fmt_v("plink", fmt, ap);
     va_end(ap);
-    fputc('\n', stderr);
-    postmsg(&cf);
-    if (logctx) {
-        log_free(logctx);
-        logctx = NULL;
-    }
-    cleanup_exit(1);
-}
-void modalfatalbox(const char *p, ...)
-{
-    struct termios cf;
-    va_list ap;
-    premsg(&cf);
-    fprintf(stderr, "FATAL ERROR: ");
-    va_start(ap, p);
-    vfprintf(stderr, p, ap);
-    va_end(ap);
-    fputc('\n', stderr);
-    postmsg(&cf);
-    if (logctx) {
-        log_free(logctx);
-        logctx = NULL;
-    }
-    cleanup_exit(1);
-}
-void nonfatal(const char *p, ...)
-{
-    struct termios cf;
-    va_list ap;
-    premsg(&cf);
-    fprintf(stderr, "ERROR: ");
-    va_start(ap, p);
-    vfprintf(stderr, p, ap);
-    va_end(ap);
-    fputc('\n', stderr);
-    postmsg(&cf);
-}
-void connection_fatal(void *frontend, const char *p, ...)
-{
-    struct termios cf;
-    va_list ap;
-    premsg(&cf);
-    fprintf(stderr, "FATAL ERROR: ");
-    va_start(ap, p);
-    vfprintf(stderr, p, ap);
-    va_end(ap);
-    fputc('\n', stderr);
-    postmsg(&cf);
-    if (logctx) {
-        log_free(logctx);
-        logctx = NULL;
-    }
-    cleanup_exit(1);
-}
-void cmdline_error(const char *p, ...)
-{
-    struct termios cf;
-    va_list ap;
-    premsg(&cf);
-    fprintf(stderr, "plink: ");
-    va_start(ap, p);
-    vfprintf(stderr, p, ap);
-    va_end(ap);
-    fputc('\n', stderr);
-    postmsg(&cf);
     exit(1);
 }
 
-static int local_tty = FALSE; /* do we have a local tty? */
+static bool local_tty = false; /* do we have a local tty? */
 
-static Backend *back;
-static void *backhandle;
-static Conf *conf;
+static Backend *backend;
+Conf *conf;
 
 /*
  * Default settings that are specific to Unix plink.
@@ -124,6 +52,11 @@ char *platform_default_s(const char *name)
     return NULL;
 }
 
+bool platform_default_b(const char *name, bool def)
+{
+    return def;
+}
+
 int platform_default_i(const char *name, int def)
 {
     return def;
@@ -146,11 +79,7 @@ char *x_get_default(const char *key)
 {
     return NULL;		       /* this is a stub */
 }
-int term_ldisc(Terminal *term, int mode)
-{
-    return FALSE;
-}
-void frontend_echoedit_update(void *frontend, int echo, int edit)
+static void plink_echoedit_update(Seat *seat, bool echo, bool edit)
 {
     /* Update stdin read mode to reflect changes in line discipline. */
     struct termios mode;
@@ -176,7 +105,7 @@ void frontend_echoedit_update(void *frontend, int echo, int edit)
 	mode.c_cc[VMIN] = 1;
 	mode.c_cc[VTIME] = 0;
 	/* FIXME: perhaps what we do with IXON/IXOFF should be an
-	 * argument to frontend_echoedit_update(), to allow
+	 * argument to the echoedit_update() method, to allow
 	 * implementation of SSH-2 "xon-xoff" and Rlogin's
 	 * equivalent? */
 	mode.c_iflag &= ~IXON;
@@ -208,7 +137,7 @@ static char *get_ttychar(struct termios *t, int index)
     return dupprintf("^<%d>", c);
 }
 
-char *get_ttymode(void *frontend, const char *mode)
+static char *plink_get_ttymode(Seat *seat, const char *mode)
 {
     /*
      * Propagate appropriate terminal modes from the local terminal,
@@ -224,7 +153,7 @@ char *get_ttymode(void *frontend, const char *mode)
 #define GET_BOOL(ourname, uxname, uxmemb, transform) \
     do { \
 	if (strcmp(mode, ourname) == 0) { \
-	    int b = (orig_termios.uxmemb & uxname) != 0; \
+	    bool b = (orig_termios.uxmemb & uxname) != 0; \
 	    transform; \
 	    return dupprintf("%d", b); \
 	} \
@@ -387,23 +316,27 @@ void cleanup_termios(void)
 }
 
 bufchain stdout_data, stderr_data;
+bufchain_sink stdout_bcs, stderr_bcs;
+StripCtrlChars *stdout_scc, *stderr_scc;
+BinarySink *stdout_bs, *stderr_bs;
+
 enum { EOF_NO, EOF_PENDING, EOF_SENT } outgoingeof;
 
-int try_output(int is_stderr)
+size_t try_output(bool is_stderr)
 {
     bufchain *chain = (is_stderr ? &stderr_data : &stdout_data);
     int fd = (is_stderr ? STDERR_FILENO : STDOUT_FILENO);
-    void *senddata;
-    int sendlen, ret;
+    ssize_t ret;
 
     if (bufchain_size(chain) > 0) {
-        int prev_nonblock = nonblock(fd);
+        bool prev_nonblock = nonblock(fd);
+        ptrlen senddata;
         do {
-            bufchain_prefix(chain, &senddata, &sendlen);
-            ret = write(fd, senddata, sendlen);
+            senddata = bufchain_prefix(chain);
+            ret = write(fd, senddata.ptr, senddata.len);
             if (ret > 0)
                 bufchain_consume(chain, ret);
-        } while (ret == sendlen && bufchain_size(chain) != 0);
+        } while (ret == senddata.len && bufchain_size(chain) != 0);
         if (!prev_nonblock)
             no_nonblock(fd);
         if (ret < 0 && errno != EAGAIN) {
@@ -418,46 +351,56 @@ int try_output(int is_stderr)
     return bufchain_size(&stdout_data) + bufchain_size(&stderr_data);
 }
 
-int from_backend(void *frontend_handle, int is_stderr,
-		 const char *data, int len)
+static size_t plink_output(
+    Seat *seat, bool is_stderr, const void *data, size_t len)
 {
-    if (is_stderr) {
-	bufchain_add(&stderr_data, data, len);
-	return try_output(TRUE);
-    } else {
-        assert(outgoingeof == EOF_NO);
-	bufchain_add(&stdout_data, data, len);
-	return try_output(FALSE);
-    }
+    assert(is_stderr || outgoingeof == EOF_NO);
+
+    BinarySink *bs = is_stderr ? stderr_bs : stdout_bs;
+    put_data(bs, data, len);
+
+    return try_output(is_stderr);
 }
 
-int from_backend_untrusted(void *frontend_handle, const char *data, int len)
-{
-    /*
-     * No "untrusted" output should get here (the way the code is
-     * currently, it's all diverted by FLAG_STDERR).
-     */
-    assert(!"Unexpected call to from_backend_untrusted()");
-    return 0; /* not reached */
-}
-
-int from_backend_eof(void *frontend_handle)
+static bool plink_eof(Seat *seat)
 {
     assert(outgoingeof == EOF_NO);
     outgoingeof = EOF_PENDING;
-    try_output(FALSE);
-    return FALSE;   /* do not respond to incoming EOF with outgoing */
+    try_output(false);
+    return false;   /* do not respond to incoming EOF with outgoing */
 }
 
-int get_userpass_input(prompts_t *p, const unsigned char *in, int inlen)
+static int plink_get_userpass_input(Seat *seat, prompts_t *p, bufchain *input)
 {
     int ret;
-    ret = cmdline_get_passwd_input(p, in, inlen);
+    ret = cmdline_get_passwd_input(p);
     if (ret == -1)
-	ret = console_get_userpass_input(p, in, inlen);
+	ret = console_get_userpass_input(p);
     return ret;
 }
 
+static const SeatVtable plink_seat_vt = {
+    plink_output,
+    plink_eof,
+    plink_get_userpass_input,
+    nullseat_notify_remote_exit,
+    console_connection_fatal,
+    nullseat_update_specials_menu,
+    plink_get_ttymode,
+    nullseat_set_busy_status,
+    console_verify_ssh_host_key,
+    console_confirm_weak_crypto_primitive,
+    console_confirm_weak_cached_hostkey,
+    nullseat_is_never_utf8,
+    plink_echoedit_update,
+    nullseat_get_x_display,
+    nullseat_get_windowid,
+    nullseat_get_window_pixel_size,
+    console_stripctrl_new,
+    console_set_trust_status,
+};
+static Seat plink_seat[1] = {{ &plink_seat_vt }};
+
 /*
  * Handle data from a local tty in PARMRK format.
  */
@@ -476,13 +419,13 @@ static void from_tty(void *vbuf, unsigned len)
 		} else {
 		    q = memchr(p, '\xff', end - p);
 		    if (q == NULL) q = end;
-		    back->send(backhandle, p, q - p);
+                    backend_send(backend, p, q - p);
 		    p = q;
 		}
 		break;
 	    case FF:
 		if (*p == '\xff') {
-		    back->send(backhandle, p, 1);
+                    backend_send(backend, p, 1);
 		    p++;
 		    state = NORMAL;
 		} else if (*p == '\0') {
@@ -492,7 +435,7 @@ static void from_tty(void *vbuf, unsigned len)
 		break;
 	    case FF00:
 		if (*p == '\0') {
-		    back->special(backhandle, TS_BRK);
+                    backend_special(backend, SS_BRK, 0);
 		} else {
 		    /* 
 		     * Pretend that PARMRK wasn't set.  This involves
@@ -511,11 +454,11 @@ static void from_tty(void *vbuf, unsigned len)
 			if (!(orig_termios.c_iflag & IGNPAR)) {
 			    /* PE/FE get passed on as NUL. */
 			    *p = 0;
-			    back->send(backhandle, p, 1);
+                            backend_send(backend, p, 1);
 			}
 		    } else {
 			/* INPCK not set.  Assume we got a parity error. */
-			back->send(backhandle, p, 1);
+                        backend_send(backend, p, 1);
 		    }
 		}
 		p++;
@@ -579,8 +522,16 @@ static void usage(void)
     printf("  -i key    private key file for user authentication\n");
     printf("  -noagent  disable use of Pageant\n");
     printf("  -agent    enable use of Pageant\n");
+    printf("  -noshare  disable use of connection sharing\n");
+    printf("  -share    enable use of connection sharing\n");
     printf("  -hostkey aa:bb:cc:...\n");
     printf("            manually specify a host key (may be repeated)\n");
+    printf("  -sanitise-stderr, -sanitise-stdout, "
+           "-no-sanitise-stderr, -no-sanitise-stdout\n");
+    printf("            do/don't strip control chars from standard "
+           "output/error\n");
+    printf("  -no-antispoof   omit anti-spoofing prompt after "
+           "authentication\n");
     printf("  -m file   read remote command(s) from file\n");
     printf("  -s        remote command is an SSH subsystem (SSH-2 only)\n");
     printf("  -N        don't start a shell/command (SSH-2 only)\n");
@@ -604,28 +555,29 @@ static void version(void)
 
 void frontend_net_error_pending(void) {}
 
-const int share_can_be_downstream = TRUE;
-const int share_can_be_upstream = TRUE;
+const bool share_can_be_downstream = true;
+const bool share_can_be_upstream = true;
 
-const int buildinfo_gtk_relevant = FALSE;
+const bool buildinfo_gtk_relevant = false;
 
 int main(int argc, char **argv)
 {
-    int sending;
-    int portnumber = -1;
+    bool sending;
     int *fdlist;
     int fd;
-    int i, fdcount, fdsize, fdstate;
+    int i, fdstate;
+    size_t fdsize;
     int exitcode;
-    int errors;
-    int use_subsystem = 0;
-    int got_host = FALSE;
-    int just_test_share_exists = FALSE;
+    bool errors;
+    enum TriState sanitise_stdout = AUTO, sanitise_stderr = AUTO;
+    bool use_subsystem = false;
+    bool just_test_share_exists = false;
     unsigned long now;
     struct winsize size;
+    const struct BackendVtable *backvt;
 
     fdlist = NULL;
-    fdcount = fdsize = 0;
+    fdsize = 0;
     /*
      * Initialise port and protocol to sensible defaults. (These
      * will be overridden by more or less anything.)
@@ -635,9 +587,18 @@ int main(int argc, char **argv)
 
     bufchain_init(&stdout_data);
     bufchain_init(&stderr_data);
+    bufchain_sink_init(&stdout_bcs, &stdout_data);
+    bufchain_sink_init(&stderr_bcs, &stderr_data);
+    stdout_bs = BinarySink_UPCAST(&stdout_bcs);
+    stderr_bs = BinarySink_UPCAST(&stderr_bcs);
     outgoingeof = EOF_NO;
 
-    flags = FLAG_STDERR | FLAG_STDERR_TTY;
+    flags = FLAG_STDERR_TTY;
+    cmdline_tooltype |=
+        (TOOLTYPE_HOST_ARG |
+         TOOLTYPE_HOST_ARG_CAN_BE_SESSION |
+         TOOLTYPE_HOST_ARG_PROTOCOL_PREFIX |
+         TOOLTYPE_HOST_ARG_FROM_LAUNCHABLE_LOAD);
 
     stderr_tty_init();
     /*
@@ -645,20 +606,20 @@ int main(int argc, char **argv)
      */
     conf = conf_new();
     do_defaults(NULL, conf);
-    loaded_session = FALSE;
+    loaded_session = false;
     default_protocol = conf_get_int(conf, CONF_protocol);
     default_port = conf_get_int(conf, CONF_port);
-    errors = 0;
+    errors = false;
     {
 	/*
 	 * Override the default protocol if PLINK_PROTOCOL is set.
 	 */
 	char *p = getenv("PLINK_PROTOCOL");
 	if (p) {
-	    const Backend *b = backend_from_name(p);
-	    if (b) {
-		default_protocol = b->protocol;
-		default_port = b->default_port;
+            const struct BackendVtable *vt = backend_vt_from_name(p);
+            if (vt) {
+                default_protocol = vt->protocol;
+                default_port = vt->default_port;
 		conf_set_int(conf, CONF_protocol, default_protocol);
 		conf_set_int(conf, CONF_port, default_port);
 	    }
@@ -666,235 +627,92 @@ int main(int argc, char **argv)
     }
     while (--argc) {
 	char *p = *++argv;
-	if (*p == '-') {
-	    int ret = cmdline_process_param(p, (argc > 1 ? argv[1] : NULL),
-					    1, conf);
-	    if (ret == -2) {
-		fprintf(stderr,
-			"plink: option \"%s\" requires an argument\n", p);
-		errors = 1;
-	    } else if (ret == 2) {
-		--argc, ++argv;
-	    } else if (ret == 1) {
-		continue;
-	    } else if (!strcmp(p, "-batch")) {
-		console_batch_mode = 1;
-	    } else if (!strcmp(p, "-s")) {
-                /* Save status to write to conf later. */
-		use_subsystem = 1;
-	    } else if (!strcmp(p, "-V") || !strcmp(p, "--version")) {
-                version();
-	    } else if (!strcmp(p, "--help")) {
-                usage();
-                exit(0);
-            } else if (!strcmp(p, "-pgpfp")) {
-                pgp_fingerprints();
-                exit(1);
-	    } else if (!strcmp(p, "-o")) {
-                if (argc <= 1) {
-                    fprintf(stderr,
-                            "plink: option \"-o\" requires an argument\n");
-		    errors = 1;
-		} else {
-                    --argc;
-		    provide_xrm_string(*++argv);
-		}
-	    } else if (!strcmp(p, "-shareexists")) {
-                just_test_share_exists = TRUE;
-	    } else if (!strcmp(p, "-fuzznet")) {
-		conf_set_int(conf, CONF_proxy_type, PROXY_FUZZ);
-		conf_set_str(conf, CONF_proxy_telnet_command,
-			     "%host");
-	    } else {
-		fprintf(stderr, "plink: unknown option \"%s\"\n", p);
-		errors = 1;
-	    }
-	} else if (*p) {
-	    if (!conf_launchable(conf) || !(got_host || loaded_session)) {
-		char *q = p;
+        int ret = cmdline_process_param(p, (argc > 1 ? argv[1] : NULL),
+                                        1, conf);
+        if (ret == -2) {
+            fprintf(stderr,
+                    "plink: option \"%s\" requires an argument\n", p);
+            errors = true;
+        } else if (ret == 2) {
+            --argc, ++argv;
+        } else if (ret == 1) {
+            continue;
+        } else if (!strcmp(p, "-batch")) {
+            console_batch_mode = true;
+        } else if (!strcmp(p, "-s")) {
+            /* Save status to write to conf later. */
+            use_subsystem = true;
+        } else if (!strcmp(p, "-V") || !strcmp(p, "--version")) {
+            version();
+        } else if (!strcmp(p, "--help")) {
+            usage();
+            exit(0);
+        } else if (!strcmp(p, "-pgpfp")) {
+            pgp_fingerprints();
+            exit(1);
+        } else if (!strcmp(p, "-o")) {
+            if (argc <= 1) {
+                fprintf(stderr,
+                        "plink: option \"-o\" requires an argument\n");
+                errors = true;
+            } else {
+                --argc;
+                /* Explicitly pass "plink" in place of appname for
+                 * error reporting purposes. appname will have been
+                 * set by be_foo.c to something more generic, probably
+                 * "PuTTY". */
+                provide_xrm_string(*++argv, "plink");
+            }
+        } else if (!strcmp(p, "-shareexists")) {
+            just_test_share_exists = true;
+        } else if (!strcmp(p, "-fuzznet")) {
+            conf_set_int(conf, CONF_proxy_type, PROXY_FUZZ);
+            conf_set_str(conf, CONF_proxy_telnet_command, "%host");
+        } else if (!strcmp(p, "-sanitise-stdout") ||
+                   !strcmp(p, "-sanitize-stdout")) {
+            sanitise_stdout = FORCE_ON;
+        } else if (!strcmp(p, "-no-sanitise-stdout") ||
+                   !strcmp(p, "-no-sanitize-stdout")) {
+            sanitise_stdout = FORCE_OFF;
+        } else if (!strcmp(p, "-sanitise-stderr") ||
+                   !strcmp(p, "-sanitize-stderr")) {
+            sanitise_stderr = FORCE_ON;
+        } else if (!strcmp(p, "-no-sanitise-stderr") ||
+                   !strcmp(p, "-no-sanitize-stderr")) {
+            sanitise_stderr = FORCE_OFF;
+        } else if (!strcmp(p, "-no-antispoof")) {
+            console_antispoof_prompt = false;
+	} else if (*p != '-') {
+            strbuf *cmdbuf = strbuf_new();
 
-		/*
-		 * If the hostname starts with "telnet:", set the
-		 * protocol to Telnet and process the string as a
-		 * Telnet URL.
-		 */
-		if (!strncmp(q, "telnet:", 7)) {
-		    char c;
+            while (argc > 0) {
+                if (cmdbuf->len > 0)
+                    put_byte(cmdbuf, ' '); /* add space separator */
+                put_datapl(cmdbuf, ptrlen_from_asciz(p));
+                if (--argc > 0)
+                    p = *++argv;
+            }
 
-		    q += 7;
-		    if (q[0] == '/' && q[1] == '/')
-			q += 2;
-		    conf_set_int(conf, CONF_protocol, PROT_TELNET);
-		    p = q;
-                    p += host_strcspn(p, ":/");
-		    c = *p;
-		    if (*p)
-			*p++ = '\0';
-		    if (c == ':')
-			conf_set_int(conf, CONF_port, atoi(p));
-		    else
-			conf_set_int(conf, CONF_port, -1);
-		    conf_set_str(conf, CONF_host, q);
-		    got_host = TRUE;
-		} else {
-		    char *r, *user, *host;
-		    /*
-		     * Before we process the [user@]host string, we
-		     * first check for the presence of a protocol
-		     * prefix (a protocol name followed by ",").
-		     */
-		    r = strchr(p, ',');
-		    if (r) {
-			const Backend *b;
-			*r = '\0';
-			b = backend_from_name(p);
-			if (b) {
-			    default_protocol = b->protocol;
-			    conf_set_int(conf, CONF_protocol,
-					 default_protocol);
-			    portnumber = b->default_port;
-			}
-			p = r + 1;
-		    }
+            conf_set_str(conf, CONF_remote_cmd, cmdbuf->s);
+            conf_set_str(conf, CONF_remote_cmd2, "");
+            conf_set_bool(conf, CONF_nopty, true);  /* command => no tty */
 
-		    /*
-		     * A nonzero length string followed by an @ is treated
-		     * as a username. (We discount an _initial_ @.) The
-		     * rest of the string (or the whole string if no @)
-		     * is treated as a session name and/or hostname.
-		     */
-		    r = strrchr(p, '@');
-		    if (r == p)
-			p++, r = NULL; /* discount initial @ */
-		    if (r) {
-			*r++ = '\0';
-			user = p, host = r;
-		    } else {
-			user = NULL, host = p;
-		    }
-
-		    /*
-		     * Now attempt to load a saved session with the
-		     * same name as the hostname.
-		     */
-		    {
-			Conf *conf2 = conf_new();
-			do_defaults(host, conf2);
-			if (loaded_session || !conf_launchable(conf2)) {
-			    /* No settings for this host; use defaults */
-			    /* (or session was already loaded with -load) */
-			    conf_set_str(conf, CONF_host, host);
-			    conf_set_int(conf, CONF_port, default_port);
-			    got_host = TRUE;
-			} else {
-			    conf_copy_into(conf, conf2);
-			    loaded_session = TRUE;
-			}
-			conf_free(conf2);
-		    }
-
-		    if (user) {
-			/* Patch in specified username. */
-			conf_set_str(conf, CONF_username, user);
-		    }
-
-		}
-	    } else {
-		char *command;
-		int cmdlen, cmdsize;
-		cmdlen = cmdsize = 0;
-		command = NULL;
-
-		while (argc) {
-		    while (*p) {
-			if (cmdlen >= cmdsize) {
-			    cmdsize = cmdlen + 512;
-			    command = sresize(command, cmdsize, char);
-			}
-			command[cmdlen++]=*p++;
-		    }
-		    if (cmdlen >= cmdsize) {
-			cmdsize = cmdlen + 512;
-			command = sresize(command, cmdsize, char);
-		    }
-		    command[cmdlen++]=' '; /* always add trailing space */
-		    if (--argc) p = *++argv;
-		}
-		if (cmdlen) command[--cmdlen]='\0';
-				       /* change trailing blank to NUL */
-		conf_set_str(conf, CONF_remote_cmd, command);
-		conf_set_str(conf, CONF_remote_cmd2, "");
-		conf_set_int(conf, CONF_nopty, TRUE);  /* command => no tty */
-
-		break;		       /* done with cmdline */
-	    }
+            strbuf_free(cmdbuf);
+            break;		       /* done with cmdline */
+        } else {
+            fprintf(stderr, "plink: unknown option \"%s\"\n", p);
+            errors = true;
 	}
     }
 
     if (errors)
 	return 1;
 
-    if (!conf_launchable(conf) || !(got_host || loaded_session)) {
+    if (!cmdline_host_ok(conf)) {
 	usage();
     }
 
-    /*
-     * Muck about with the hostname in various ways.
-     */
-    {
-	char *hostbuf = dupstr(conf_get_str(conf, CONF_host));
-	char *host = hostbuf;
-	char *p, *q;
-
-	/*
-	 * Trim leading whitespace.
-	 */
-	host += strspn(host, " \t");
-
-	/*
-	 * See if host is of the form user@host, and separate out
-	 * the username if so.
-	 */
-	if (host[0] != '\0') {
-	    char *atsign = strrchr(host, '@');
-	    if (atsign) {
-		*atsign = '\0';
-		conf_set_str(conf, CONF_username, host);
-		host = atsign + 1;
-	    }
-	}
-
-        /*
-         * Trim a colon suffix off the hostname if it's there. In
-         * order to protect unbracketed IPv6 address literals
-         * against this treatment, we do not do this if there's
-         * _more_ than one colon.
-         */
-        {
-            char *c = host_strchr(host, ':');
- 
-            if (c) {
-                char *d = host_strchr(c+1, ':');
-                if (!d)
-                    *c = '\0';
-            }
-        }
-
-	/*
-	 * Remove any remaining whitespace.
-	 */
-	p = hostbuf;
-	q = host;
-	while (*q) {
-	    if (*q != ' ' && *q != '\t')
-		*p++ = *q;
-	    q++;
-	}
-	*p = '\0';
-
-	conf_set_str(conf, CONF_host, hostbuf);
-	sfree(hostbuf);
-    }
+    prepare_session(conf);
 
     /*
      * Perform command-line overrides on session configuration.
@@ -917,7 +735,7 @@ int main(int argc, char **argv)
      * Apply subsystem status.
      */
     if (use_subsystem)
-        conf_set_int(conf, CONF_ssh_subsys, TRUE);
+        conf_set_bool(conf, CONF_ssh_subsys, true);
 
     if (!*conf_get_str(conf, CONF_remote_cmd) &&
 	!*conf_get_str(conf, CONF_remote_cmd2) &&
@@ -928,19 +746,13 @@ int main(int argc, char **argv)
      * Select protocol. This is farmed out into a table in a
      * separate file to enable an ssh-free variant.
      */
-    back = backend_from_proto(conf_get_int(conf, CONF_protocol));
-    if (back == NULL) {
+    backvt = backend_vt_from_proto(conf_get_int(conf, CONF_protocol));
+    if (!backvt) {
 	fprintf(stderr,
 		"Internal fault: Unsupported protocol found\n");
 	return 1;
     }
 
-    /*
-     * Select port.
-     */
-    if (portnumber != -1)
-	conf_set_int(conf, CONF_port, portnumber);
-
     /*
      * Block SIGPIPE, so that we'll get EPIPE individually on
      * particular network connections that go wrong.
@@ -970,6 +782,33 @@ int main(int argc, char **argv)
 	conf_set_int(conf, CONF_height, size.ws_row);
     }
 
+    /*
+     * Decide whether to sanitise control sequences out of standard
+     * output and standard error.
+     *
+     * If we weren't given a command-line override, we do this if (a)
+     * the fd in question is pointing at a terminal, and (b) we aren't
+     * trying to allocate a terminal as part of the session.
+     *
+     * (Rationale: the risk of control sequences is that they cause
+     * confusion when sent to a local terminal, so if there isn't one,
+     * no problem. Also, if we allocate a remote terminal, then we
+     * sent a terminal type, i.e. we told it what kind of escape
+     * sequences we _like_, i.e. we were expecting to receive some.)
+     */
+    if (sanitise_stdout == FORCE_ON ||
+        (sanitise_stdout == AUTO && isatty(STDOUT_FILENO) &&
+         conf_get_bool(conf, CONF_nopty))) {
+        stdout_scc = stripctrl_new(stdout_bs, true, L'\0');
+        stdout_bs = BinarySink_UPCAST(stdout_scc);
+    }
+    if (sanitise_stderr == FORCE_ON ||
+        (sanitise_stderr == AUTO && isatty(STDERR_FILENO) &&
+         conf_get_bool(conf, CONF_nopty))) {
+        stderr_scc = stripctrl_new(stderr_bs, true, L'\0');
+        stderr_bs = BinarySink_UPCAST(stderr_scc);
+    }
+
     sk_init();
     uxsel_init();
 
@@ -979,19 +818,19 @@ int main(int argc, char **argv)
      * the "simple" flag.
      */
     if (conf_get_int(conf, CONF_protocol) == PROT_SSH &&
-	!conf_get_int(conf, CONF_x11_forward) &&
-	!conf_get_int(conf, CONF_agentfwd) &&
+	!conf_get_bool(conf, CONF_x11_forward) &&
+	!conf_get_bool(conf, CONF_agentfwd) &&
 	!conf_get_str_nthstrkey(conf, CONF_portfwd, 0))
-	conf_set_int(conf, CONF_ssh_simple, TRUE);
+	conf_set_bool(conf, CONF_ssh_simple, true);
 
     if (just_test_share_exists) {
-        if (!back->test_for_upstream) {
+        if (!backvt->test_for_upstream) {
             fprintf(stderr, "Connection sharing not supported for connection "
-                    "type '%s'\n", back->name);
+                    "type '%s'\n", backvt->name);
             return 1;
         }
-        if (back->test_for_upstream(conf_get_str(conf, CONF_host),
-                                    conf_get_int(conf, CONF_port), conf))
+        if (backvt->test_for_upstream(conf_get_str(conf, CONF_host),
+                                      conf_get_int(conf, CONF_port), conf))
             return 0;
         else
             return 1;
@@ -1000,30 +839,28 @@ int main(int argc, char **argv)
     /*
      * Start up the connection.
      */
-    logctx = log_init(NULL, conf);
-    console_provide_logctx(logctx);
+    logctx = log_init(default_logpolicy, conf);
     {
 	const char *error;
 	char *realhost;
 	/* nodelay is only useful if stdin is a terminal device */
-	int nodelay = conf_get_int(conf, CONF_tcp_nodelay) && isatty(0);
+	bool nodelay = conf_get_bool(conf, CONF_tcp_nodelay) && isatty(0);
 
 	/* This is a good place for a fuzzer to fork us. */
 #ifdef __AFL_HAVE_MANUAL_CONTROL
 	__AFL_INIT();
 #endif
 
-	error = back->init(NULL, &backhandle, conf,
-			   conf_get_str(conf, CONF_host),
-			   conf_get_int(conf, CONF_port),
-			   &realhost, nodelay,
-			   conf_get_int(conf, CONF_tcp_keepalives));
+        error = backend_init(backvt, plink_seat, &backend, logctx, conf,
+                             conf_get_str(conf, CONF_host),
+                             conf_get_int(conf, CONF_port),
+                             &realhost, nodelay,
+                             conf_get_bool(conf, CONF_tcp_keepalives));
 	if (error) {
 	    fprintf(stderr, "Unable to open connection:\n%s\n", error);
 	    return 1;
 	}
-	back->provide_logctx(backhandle, logctx);
-	ldisc_create(conf, NULL, back, backhandle, NULL);
+        ldisc_create(conf, NULL, backend, plink_seat);
 	sfree(realhost);
     }
 
@@ -1034,40 +871,37 @@ int main(int argc, char **argv)
      */
     local_tty = (tcgetattr(STDIN_FILENO, &orig_termios) == 0);
     atexit(cleanup_termios);
-    frontend_echoedit_update(NULL, 1, 1);
-    sending = FALSE;
+    seat_echoedit_update(plink_seat, 1, 1);
+    sending = false;
     now = GETTICKCOUNT();
 
+    pollwrapper *pw = pollwrap_new();
+
     while (1) {
-	fd_set rset, wset, xset;
-	int maxfd;
 	int rwx;
 	int ret;
         unsigned long next;
 
-	FD_ZERO(&rset);
-	FD_ZERO(&wset);
-	FD_ZERO(&xset);
-	maxfd = 0;
+        pollwrap_clear(pw);
 
-	FD_SET_MAX(signalpipe[0], maxfd, rset);
+	pollwrap_add_fd_rwx(pw, signalpipe[0], SELECT_R);
 
 	if (!sending &&
-	    back->connected(backhandle) &&
-	    back->sendok(backhandle) &&
-	    back->sendbuffer(backhandle) < MAX_STDIN_BACKLOG) {
+            backend_connected(backend) &&
+            backend_sendok(backend) &&
+            backend_sendbuffer(backend) < MAX_STDIN_BACKLOG) {
 	    /* If we're OK to send, then try to read from stdin. */
-	    FD_SET_MAX(STDIN_FILENO, maxfd, rset);
+            pollwrap_add_fd_rwx(pw, STDIN_FILENO, SELECT_R);
 	}
 
 	if (bufchain_size(&stdout_data) > 0) {
 	    /* If we have data for stdout, try to write to stdout. */
-	    FD_SET_MAX(STDOUT_FILENO, maxfd, wset);
+            pollwrap_add_fd_rwx(pw, STDOUT_FILENO, SELECT_W);
 	}
 
 	if (bufchain_size(&stderr_data) > 0) {
 	    /* If we have data for stderr, try to write to stderr. */
-	    FD_SET_MAX(STDERR_FILENO, maxfd, wset);
+            pollwrap_add_fd_rwx(pw, STDERR_FILENO, SELECT_W);
 	}
 
 	/* Count the currently active fds. */
@@ -1076,37 +910,25 @@ int main(int argc, char **argv)
 	     fd = next_fd(&fdstate, &rwx)) i++;
 
 	/* Expand the fdlist buffer if necessary. */
-	if (i > fdsize) {
-	    fdsize = i + 16;
-	    fdlist = sresize(fdlist, fdsize, int);
-	}
+        sgrowarray(fdlist, fdsize, i);
 
 	/*
-	 * Add all currently open fds to the select sets, and store
-	 * them in fdlist as well.
+	 * Add all currently open fds to pw, and store them in fdlist
+	 * as well.
 	 */
-	fdcount = 0;
+	int fdcount = 0;
 	for (fd = first_fd(&fdstate, &rwx); fd >= 0;
 	     fd = next_fd(&fdstate, &rwx)) {
 	    fdlist[fdcount++] = fd;
-	    if (rwx & 1)
-		FD_SET_MAX(fd, maxfd, rset);
-	    if (rwx & 2)
-		FD_SET_MAX(fd, maxfd, wset);
-	    if (rwx & 4)
-		FD_SET_MAX(fd, maxfd, xset);
+            pollwrap_add_fd_rwx(pw, fd, rwx);
 	}
 
         if (toplevel_callback_pending()) {
-            struct timeval tv;
-            tv.tv_sec = 0;
-            tv.tv_usec = 0;
-            ret = select(maxfd, &rset, &wset, &xset, &tv);
+            ret = pollwrap_poll_instant(pw);
         } else if (run_timers(now, &next)) {
             do {
                 unsigned long then;
                 long ticks;
-                struct timeval tv;
 
 		then = now;
 		now = GETTICKCOUNT();
@@ -1114,88 +936,95 @@ int main(int argc, char **argv)
 		    ticks = 0;
 		else
 		    ticks = next - now;
-		tv.tv_sec = ticks / 1000;
-		tv.tv_usec = ticks % 1000 * 1000;
-                ret = select(maxfd, &rset, &wset, &xset, &tv);
-                if (ret == 0)
+
+                bool overflow = false;
+                if (ticks > INT_MAX) {
+                    ticks = INT_MAX;
+                    overflow = true;
+                }
+
+                ret = pollwrap_poll_timeout(pw, ticks);
+                if (ret == 0 && !overflow)
                     now = next;
                 else
                     now = GETTICKCOUNT();
             } while (ret < 0 && errno == EINTR);
         } else {
-            ret = select(maxfd, &rset, &wset, &xset, NULL);
+            ret = pollwrap_poll_endless(pw);
         }
 
         if (ret < 0 && errno == EINTR)
             continue;
 
 	if (ret < 0) {
-	    perror("select");
+	    perror("poll");
 	    exit(1);
 	}
 
 	for (i = 0; i < fdcount; i++) {
 	    fd = fdlist[i];
+            int rwx = pollwrap_get_fd_rwx(pw, fd);
             /*
              * We must process exceptional notifications before
              * ordinary readability ones, or we may go straight
              * past the urgent marker.
              */
-	    if (FD_ISSET(fd, &xset))
-		select_result(fd, 4);
-	    if (FD_ISSET(fd, &rset))
-		select_result(fd, 1);
-	    if (FD_ISSET(fd, &wset))
-		select_result(fd, 2);
+	    if (rwx & SELECT_X)
+		select_result(fd, SELECT_X);
+	    if (rwx & SELECT_R)
+		select_result(fd, SELECT_R);
+	    if (rwx & SELECT_W)
+		select_result(fd, SELECT_W);
 	}
 
-	if (FD_ISSET(signalpipe[0], &rset)) {
+	if (pollwrap_check_fd_rwx(pw, signalpipe[0], SELECT_R)) {
 	    char c[1];
 	    struct winsize size;
 	    if (read(signalpipe[0], c, 1) <= 0)
 		/* ignore error */;
 	    /* ignore its value; it'll be `x' */
 	    if (ioctl(STDIN_FILENO, TIOCGWINSZ, (void *)&size) >= 0)
-		back->size(backhandle, size.ws_col, size.ws_row);
+                backend_size(backend, size.ws_col, size.ws_row);
 	}
 
-	if (FD_ISSET(STDIN_FILENO, &rset)) {
+	if (pollwrap_check_fd_rwx(pw, STDIN_FILENO, SELECT_R)) {
 	    char buf[4096];
 	    int ret;
 
-	    if (back->connected(backhandle)) {
+            if (backend_connected(backend)) {
 		ret = read(STDIN_FILENO, buf, sizeof(buf));
+                noise_ultralight(NOISE_SOURCE_IOLEN, ret);
 		if (ret < 0) {
 		    perror("stdin: read");
 		    exit(1);
 		} else if (ret == 0) {
-		    back->special(backhandle, TS_EOF);
-		    sending = FALSE;   /* send nothing further after this */
+                    backend_special(backend, SS_EOF, 0);
+		    sending = false;   /* send nothing further after this */
 		} else {
 		    if (local_tty)
 			from_tty(buf, ret);
 		    else
-			back->send(backhandle, buf, ret);
+                        backend_send(backend, buf, ret);
 		}
 	    }
 	}
 
-	if (FD_ISSET(STDOUT_FILENO, &wset)) {
-	    back->unthrottle(backhandle, try_output(FALSE));
+	if (pollwrap_check_fd_rwx(pw, STDOUT_FILENO, SELECT_W)) {
+            backend_unthrottle(backend, try_output(false));
 	}
 
-	if (FD_ISSET(STDERR_FILENO, &wset)) {
-	    back->unthrottle(backhandle, try_output(TRUE));
+	if (pollwrap_check_fd_rwx(pw, STDERR_FILENO, SELECT_W)) {
+            backend_unthrottle(backend, try_output(true));
 	}
 
         run_toplevel_callbacks();
 
-	if (!back->connected(backhandle) &&
+        if (!backend_connected(backend) &&
 	    bufchain_size(&stdout_data) == 0 &&
 	    bufchain_size(&stderr_data) == 0)
 	    break;		       /* we closed the connection */
     }
-    exitcode = back->exitcode(backhandle);
+    exitcode = backend_exitcode(backend);
     if (exitcode < 0) {
 	fprintf(stderr, "Remote process exit code unavailable\n");
 	exitcode = 1;		       /* this is an error condition */
