@@ -25,12 +25,13 @@ printer_job *printer_start_job(char *printer)
     return ret;
 }
 
-void printer_job_data(printer_job *pj, void *data, int len)
+void printer_job_data(printer_job *pj, const void *data, size_t len)
 {
     if (!pj)
 	return;
 
-    fwrite(data, 1, len, pj->fp);
+    if (fwrite(data, 1, len, pj->fp) < len)
+	/* ignore */;
 }
 
 void printer_finish_job(printer_job *pj)
