@@ -12,15 +12,14 @@
 #include <utime.h>
 #include <errno.h>
 #include <assert.h>
-#include <glob.h>
-#ifndef HAVE_NO_SYS_SELECT_H
-#include <sys/select.h>
-#endif
 
 #include "putty.h"
 #include "ssh.h"
 #include "psftp.h"
-#include "int64.h"
+
+#if HAVE_GLOB_H
+#include <glob.h>
+#endif
 
 /*
  * In PSFTP our selects are synchronous, so these functions are
@@ -38,7 +37,7 @@ void platform_get_x11_auth(struct X11Display *display, Conf *conf)
 {
     /* Do nothing, therefore no auth. */
 }
-const int platform_uses_x11_unix_by_default = TRUE;
+const bool platform_uses_x11_unix_by_default = true;
 
 /*
  * Default settings that are specific to PSFTP.
@@ -48,6 +47,11 @@ char *platform_default_s(const char *name)
     return NULL;
 }
 
+bool platform_default_b(const char *name, bool def)
+{
+    return def;
+}
+
 int platform_default_i(const char *name, int def)
 {
     return def;
@@ -66,14 +70,12 @@ Filename *platform_default_filename(const char *name)
 	return filename_from_str("");
 }
 
-char *get_ttymode(void *frontend, const char *mode) { return NULL; }
-
-int get_userpass_input(prompts_t *p, const unsigned char *in, int inlen)
+int filexfer_get_userpass_input(Seat *seat, prompts_t *p, bufchain *input)
 {
     int ret;
-    ret = cmdline_get_passwd_input(p, in, inlen);
+    ret = cmdline_get_passwd_input(p);
     if (ret == -1)
-	ret = console_get_userpass_input(p, in, inlen);
+	ret = console_get_userpass_input(p);
     return ret;
 }
 
@@ -96,7 +98,7 @@ char *psftp_lcd(char *dir)
 char *psftp_getcwd(void)
 {
     char *buffer, *ret;
-    int size = 256;
+    size_t size = 256;
 
     buffer = snewn(size, char);
     while (1) {
@@ -111,8 +113,7 @@ char *psftp_getcwd(void)
 	 * Otherwise, ERANGE was returned, meaning the buffer
 	 * wasn't big enough.
 	 */
-	size = size * 3 / 2;
-	buffer = sresize(buffer, size, char);
+        sgrowarray(buffer, size, size);
     }
 }
 
@@ -120,7 +121,7 @@ struct RFile {
     int fd;
 };
 
-RFile *open_existing_file(const char *name, uint64 *size,
+RFile *open_existing_file(const char *name, uint64_t *size,
 			  unsigned long *mtime, unsigned long *atime,
                           long *perms)
 {
@@ -142,8 +143,7 @@ RFile *open_existing_file(const char *name, uint64 *size,
 	}
 
 	if (size)
-	    *size = uint64_make((statbuf.st_size >> 16) >> 16,
-				statbuf.st_size);
+	    *size = statbuf.st_size;
 	 	
 	if (mtime)
 	    *mtime = statbuf.st_mtime;
@@ -192,7 +192,7 @@ WFile *open_new_file(const char *name, long perms)
 }
 
 
-WFile *open_existing_wfile(const char *name, uint64 *size)
+WFile *open_existing_wfile(const char *name, uint64_t *size)
 {
     int fd;
     WFile *ret;
@@ -212,8 +212,7 @@ WFile *open_existing_wfile(const char *name, uint64 *size)
 	    memset(&statbuf, 0, sizeof(statbuf));
 	}
 
-	*size = uint64_make((statbuf.st_size >> 16) >> 16,
-			    statbuf.st_size);
+	*size = statbuf.st_size;
     }
 
     return ret;
@@ -262,13 +261,10 @@ void close_wfile(WFile *f)
 
 /* Seek offset bytes through file, from whence, where whence is
    FROM_START, FROM_CURRENT, or FROM_END */
-int seek_file(WFile *f, uint64 offset, int whence)
+int seek_file(WFile *f, uint64_t offset, int whence)
 {
-    off_t fileofft;
     int lseek_whence;
     
-    fileofft = (((off_t) offset.hi << 16) << 16) + offset.lo;
-
     switch (whence) {
     case FROM_START:
 	lseek_whence = SEEK_SET;
@@ -283,19 +279,12 @@ int seek_file(WFile *f, uint64 offset, int whence)
 	return -1;
     }
 
-    return lseek(f->fd, fileofft, lseek_whence) >= 0 ? 0 : -1;
+    return lseek(f->fd, offset, lseek_whence) >= 0 ? 0 : -1;
 }
 
-uint64 get_file_posn(WFile *f)
+uint64_t get_file_posn(WFile *f)
 {
-    off_t fileofft;
-    uint64 ret;
-
-    fileofft = lseek(f->fd, (off_t) 0, SEEK_CUR);
-
-    ret = uint64_make((fileofft >> 16) >> 16, fileofft);
-
-    return ret;
+    return lseek(f->fd, (off_t) 0, SEEK_CUR);
 }
 
 int file_type(const char *name)
@@ -321,14 +310,16 @@ struct DirHandle {
     DIR *dir;
 };
 
-DirHandle *open_directory(const char *name)
+DirHandle *open_directory(const char *name, const char **errmsg)
 {
     DIR *dir;
     DirHandle *ret;
 
     dir = opendir(name);
-    if (!dir)
+    if (!dir) {
+        *errmsg = strerror(errno);
 	return NULL;
+    }
 
     ret = snew(DirHandle);
     ret->dir = dir;
@@ -356,7 +347,7 @@ void close_directory(DirHandle *dir)
     sfree(dir);
 }
 
-int test_wildcard(const char *name, int cmdline)
+int test_wildcard(const char *name, bool cmdline)
 {
     struct stat statbuf;
 
@@ -370,6 +361,7 @@ int test_wildcard(const char *name, int cmdline)
 	 */
 	return WCTYPE_NONEXISTENT;
     } else {
+#if HAVE_GLOB_H
 	glob_t globbed;
 	int ret = WCTYPE_NONEXISTENT;
 
@@ -380,12 +372,18 @@ int test_wildcard(const char *name, int cmdline)
 	}
 
 	return ret;
+#else
+        /* On a system without glob.h, we just have to return a
+         * failure code */
+        return WCTYPE_NONEXISTENT;
+#endif
     }
 }
 
 /*
  * Actually return matching file names for a local wildcard.
  */
+#if HAVE_GLOB_H
 struct WildcardMatcher {
     glob_t globbed;
     int i;
@@ -412,8 +410,22 @@ void finish_wildcard_matching(WildcardMatcher *dir) {
     globfree(&dir->globbed);
     sfree(dir);
 }
+#else
+WildcardMatcher *begin_wildcard_matching(const char *name)
+{
+    return NULL;
+}
+char *wildcard_get_filename(WildcardMatcher *dir)
+{
+    unreachable("Can't construct a valid WildcardMatcher without <glob.h>");
+}
+void finish_wildcard_matching(WildcardMatcher *dir)
+{
+    unreachable("Can't construct a valid WildcardMatcher without <glob.h>");
+}
+#endif
 
-char *stripslashes(const char *str, int local)
+char *stripslashes(const char *str, bool local)
 {
     char *p;
 
@@ -427,41 +439,47 @@ char *stripslashes(const char *str, int local)
     return (char *)str;
 }
 
-int vet_filename(const char *name)
+bool vet_filename(const char *name)
 {
     if (strchr(name, '/'))
-	return FALSE;
+	return false;
 
     if (name[0] == '.' && (!name[1] || (name[1] == '.' && !name[2])))
-	return FALSE;
+	return false;
 
-    return TRUE;
+    return true;
 }
 
-int create_directory(const char *name)
+bool create_directory(const char *name)
 {
     return mkdir(name, 0777) == 0;
 }
 
 char *dir_file_cat(const char *dir, const char *file)
 {
-    return dupcat(dir, "/", file, NULL);
+    ptrlen dir_pl = ptrlen_from_asciz(dir);
+    return dupcat(
+        dir, ptrlen_endswith(dir_pl, PTRLEN_LITERAL("/"), NULL) ? "" : "/",
+        file, NULL);
 }
 
 /*
  * Do a select() between all currently active network fds and
  * optionally stdin.
  */
-static int ssh_sftp_do_select(int include_stdin, int no_fds_ok)
+static int ssh_sftp_do_select(bool include_stdin, bool no_fds_ok)
 {
-    fd_set rset, wset, xset;
-    int i, fdcount, fdsize, *fdlist;
-    int fd, fdstate, rwx, ret, maxfd;
+    int i, *fdlist;
+    size_t fdsize;
+    int fd, fdcount, fdstate, rwx, ret;
     unsigned long now = GETTICKCOUNT();
     unsigned long next;
+    bool done_something = false;
 
     fdlist = NULL;
-    fdcount = fdsize = 0;
+    fdsize = 0;
+
+    pollwrapper *pw = pollwrap_new();
 
     do {
 
@@ -470,19 +488,15 @@ static int ssh_sftp_do_select(int include_stdin, int no_fds_ok)
 	for (fd = first_fd(&fdstate, &rwx); fd >= 0;
 	     fd = next_fd(&fdstate, &rwx)) i++;
 
-	if (i < 1 && !no_fds_ok)
+	if (i < 1 && !no_fds_ok && !toplevel_callback_pending()) {
+            pollwrap_free(pw);
 	    return -1;		       /* doom */
+        }
 
 	/* Expand the fdlist buffer if necessary. */
-	if (i > fdsize) {
-	    fdsize = i + 16;
-	    fdlist = sresize(fdlist, fdsize, int);
-	}
+        sgrowarray(fdlist, fdsize, i);
 
-	FD_ZERO(&rset);
-	FD_ZERO(&wset);
-	FD_ZERO(&xset);
-	maxfd = 0;
+        pollwrap_clear(pw);
 
 	/*
 	 * Add all currently open fds to the select sets, and store
@@ -492,29 +506,20 @@ static int ssh_sftp_do_select(int include_stdin, int no_fds_ok)
 	for (fd = first_fd(&fdstate, &rwx); fd >= 0;
 	     fd = next_fd(&fdstate, &rwx)) {
 	    fdlist[fdcount++] = fd;
-	    if (rwx & 1)
-		FD_SET_MAX(fd, maxfd, rset);
-	    if (rwx & 2)
-		FD_SET_MAX(fd, maxfd, wset);
-	    if (rwx & 4)
-		FD_SET_MAX(fd, maxfd, xset);
+            pollwrap_add_fd_rwx(pw, fd, rwx);
 	}
 
 	if (include_stdin)
-	    FD_SET_MAX(0, maxfd, rset);
+	    pollwrap_add_fd_rwx(pw, 0, SELECT_R);
 
         if (toplevel_callback_pending()) {
-            struct timeval tv;
-            tv.tv_sec = 0;
-            tv.tv_usec = 0;
-            ret = select(maxfd, &rset, &wset, &xset, &tv);
+            ret = pollwrap_poll_instant(pw);
             if (ret == 0)
-                run_toplevel_callbacks();
+                done_something |= run_toplevel_callbacks();
         } else if (run_timers(now, &next)) {
             do {
                 unsigned long then;
                 long ticks;
-                struct timeval tv;
 
 		then = now;
 		now = GETTICKCOUNT();
@@ -522,46 +527,54 @@ static int ssh_sftp_do_select(int include_stdin, int no_fds_ok)
 		    ticks = 0;
 		else
 		    ticks = next - now;
-		tv.tv_sec = ticks / 1000;
-		tv.tv_usec = ticks % 1000 * 1000;
-                ret = select(maxfd, &rset, &wset, &xset, &tv);
-                if (ret == 0)
+
+                bool overflow = false;
+                if (ticks > INT_MAX) {
+                    ticks = INT_MAX;
+                    overflow = true;
+                }
+
+                ret = pollwrap_poll_timeout(pw, ticks);
+                if (ret == 0 && !overflow)
                     now = next;
                 else
                     now = GETTICKCOUNT();
             } while (ret < 0 && errno == EINTR);
         } else {
             do {
-                ret = select(maxfd, &rset, &wset, &xset, NULL);
+                ret = pollwrap_poll_endless(pw);
             } while (ret < 0 && errno == EINTR);
         }
-    } while (ret == 0);
+    } while (ret == 0 && !done_something);
 
     if (ret < 0) {
-	perror("select");
+	perror("poll");
 	exit(1);
     }
 
     for (i = 0; i < fdcount; i++) {
 	fd = fdlist[i];
+        int rwx = pollwrap_get_fd_rwx(pw, fd);
 	/*
 	 * We must process exceptional notifications before
 	 * ordinary readability ones, or we may go straight
 	 * past the urgent marker.
 	 */
-	if (FD_ISSET(fd, &xset))
-	    select_result(fd, 4);
-	if (FD_ISSET(fd, &rset))
-	    select_result(fd, 1);
-	if (FD_ISSET(fd, &wset))
-	    select_result(fd, 2);
+	if (rwx & SELECT_X)
+	    select_result(fd, SELECT_X);
+	if (rwx & SELECT_R)
+	    select_result(fd, SELECT_R);
+	if (rwx & SELECT_W)
+	    select_result(fd, SELECT_W);
     }
 
     sfree(fdlist);
 
     run_toplevel_callbacks();
 
-    return FD_ISSET(0, &rset) ? 1 : 0;
+    int toret = pollwrap_check_fd_rwx(pw, 0, SELECT_R) ? 1 : 0;
+    pollwrap_free(pw);
+    return toret;
 }
 
 /*
@@ -569,16 +582,17 @@ static int ssh_sftp_do_select(int include_stdin, int no_fds_ok)
  */
 int ssh_sftp_loop_iteration(void)
 {
-    return ssh_sftp_do_select(FALSE, FALSE);
+    return ssh_sftp_do_select(false, false);
 }
 
 /*
  * Read a PSFTP command line from stdin.
  */
-char *ssh_sftp_get_cmdline(const char *prompt, int no_fds_ok)
+char *ssh_sftp_get_cmdline(const char *prompt, bool no_fds_ok)
 {
     char *buf;
-    int buflen, bufsize, ret;
+    size_t buflen, bufsize;
+    int ret;
 
     fputs(prompt, stdout);
     fflush(stdout);
@@ -587,17 +601,14 @@ char *ssh_sftp_get_cmdline(const char *prompt, int no_fds_ok)
     buflen = bufsize = 0;
 
     while (1) {
-	ret = ssh_sftp_do_select(TRUE, no_fds_ok);
+	ret = ssh_sftp_do_select(true, no_fds_ok);
 	if (ret < 0) {
 	    printf("connection died\n");
             sfree(buf);
 	    return NULL;	       /* woop woop */
 	}
 	if (ret > 0) {
-	    if (buflen >= bufsize) {
-		bufsize = buflen + 512;
-		buf = sresize(buf, bufsize, char);
-	    }
+            sgrowarray(buf, bufsize, buflen);
 	    ret = read(0, buf+buflen, 1);
 	    if (ret < 0) {
 		perror("read");
@@ -622,7 +633,7 @@ void frontend_net_error_pending(void) {}
 
 void platform_psftp_pre_conn_setup(void) {}
 
-const int buildinfo_gtk_relevant = FALSE;
+const bool buildinfo_gtk_relevant = false;
 
 /*
  * Main program: do platform-specific initialisation and then call
