@@ -69,12 +69,12 @@ struct interval {
 };
 
 /* auxiliary function for binary search in interval table */
-static int bisearch(unsigned int ucs, const struct interval *table, int max) {
+static bool bisearch(unsigned int ucs, const struct interval *table, int max) {
   int min = 0;
   int mid;
 
   if (ucs < table[0].first || ucs > table[max].last)
-    return 0;
+    return false;
   while (max >= min) {
     mid = (min + max) / 2;
     if (ucs > table[mid].last)
@@ -82,10 +82,10 @@ static int bisearch(unsigned int ucs, const struct interval *table, int max) {
     else if (ucs < table[mid].first)
       max = mid - 1;
     else
-      return 1;
+      return true;
   }
 
-  return 0;
+  return false;
 }
 
 
