@@ -11,9 +11,7 @@
 #include "storage.h"
 #include "ssh.h"
 
-int console_batch_mode = FALSE;
-
-static void *console_logctx = NULL;
+bool console_batch_mode = false;
 
 /*
  * Clean up and exit.
@@ -26,28 +24,69 @@ void cleanup_exit(int code)
     sk_cleanup();
 
     random_save_seed();
-#ifdef MSCRYPTOAPI
-    crypto_wrapup();
-#endif
 
     exit(code);
 }
 
-void set_busy_status(void *frontend, int status)
+/*
+ * Various error message and/or fatal exit functions.
+ */
+void console_print_error_msg(const char *prefix, const char *msg)
 {
+    fputs(prefix, stderr);
+    fputs(": ", stderr);
+    fputs(msg, stderr);
+    fputc('\n', stderr);
+    fflush(stderr);
 }
 
-void notify_remote_exit(void *frontend)
+void console_print_error_msg_fmt_v(
+    const char *prefix, const char *fmt, va_list ap)
 {
+    char *msg = dupvprintf(fmt, ap);
+    console_print_error_msg(prefix, msg);
+    sfree(msg);
+}
+
+void console_print_error_msg_fmt(const char *prefix, const char *fmt, ...)
+{
+    va_list ap;
+    va_start(ap, fmt);
+    console_print_error_msg_fmt_v(prefix, fmt, ap);
+    va_end(ap);
+}
+
+void modalfatalbox(const char *fmt, ...)
+{
+    va_list ap;
+    va_start(ap, fmt);
+    console_print_error_msg_fmt_v("FATAL ERROR", fmt, ap);
+    va_end(ap);
+    cleanup_exit(1);
+}
+
+void nonfatal(const char *fmt, ...)
+{
+    va_list ap;
+    va_start(ap, fmt);
+    console_print_error_msg_fmt_v("ERROR", fmt, ap);
+    va_end(ap);
+}
+
+void console_connection_fatal(Seat *seat, const char *msg)
+{
+    console_print_error_msg("FATAL ERROR", msg);
+    cleanup_exit(1);
 }
 
 void timer_change_notify(unsigned long next)
 {
 }
 
-int verify_ssh_host_key(void *frontend, char *host, int port,
-                        const char *keytype, char *keystr, char *fingerprint,
-                        void (*callback)(void *ctx, int result), void *ctx)
+int console_verify_ssh_host_key(
+    Seat *seat, const char *host, int port,
+    const char *keytype, char *keystr, char *fingerprint,
+    void (*callback)(void *ctx, int result), void *ctx)
 {
     int ret;
     HANDLE hin;
@@ -150,16 +189,9 @@ int verify_ssh_host_key(void *frontend, char *host, int port,
     }
 }
 
-void update_specials_menu(void *frontend)
-{
-}
-
-/*
- * Ask whether the selected algorithm is acceptable (since it was
- * below the configured 'warn' threshold).
- */
-int askalg(void *frontend, const char *algtype, const char *algname,
-	   void (*callback)(void *ctx, int result), void *ctx)
+int console_confirm_weak_crypto_primitive(
+    Seat *seat, const char *algtype, const char *algname,
+    void (*callback)(void *ctx, int result), void *ctx)
 {
     HANDLE hin;
     DWORD savemode, i;
@@ -199,8 +231,9 @@ int askalg(void *frontend, const char *algtype, const char *algname,
     }
 }
 
-int askhk(void *frontend, const char *algname, const char *betteralgs,
-          void (*callback)(void *ctx, int result), void *ctx)
+int console_confirm_weak_cached_hostkey(
+    Seat *seat, const char *algname, const char *betteralgs,
+    void (*callback)(void *ctx, int result), void *ctx)
 {
     HANDLE hin;
     DWORD savemode, i;
@@ -246,12 +279,41 @@ int askhk(void *frontend, const char *algname, const char *betteralgs,
     }
 }
 
+bool is_interactive(void)
+{
+    return is_console_handle(GetStdHandle(STD_INPUT_HANDLE));
+}
+
+bool console_antispoof_prompt = true;
+bool console_set_trust_status(Seat *seat, bool trusted)
+{
+    if (console_batch_mode || !is_interactive() || !console_antispoof_prompt) {
+        /*
+         * In batch mode, we don't need to worry about the server
+         * mimicking our interactive authentication, because the user
+         * already knows not to expect any.
+         *
+         * If standard input isn't connected to a terminal, likewise,
+         * because even if the server did send a spoof authentication
+         * prompt, the user couldn't respond to it via the terminal
+         * anyway.
+         *
+         * We also vacuously return success if the user has purposely
+         * disabled the antispoof prompt.
+         */
+        return true;
+    }
+
+    return false;
+}
+
 /*
  * Ask whether to wipe a session log file before writing to it.
  * Returns 2 for wipe, 1 for append, 0 for cancel (don't log).
  */
-int askappend(void *frontend, Filename *filename,
-	      void (*callback)(void *ctx, int result), void *ctx)
+static int console_askappend(LogPolicy *lp, Filename *filename,
+                             void (*callback)(void *ctx, int result),
+                             void *ctx)
 {
     HANDLE hin;
     DWORD savemode, i;
@@ -330,35 +392,45 @@ void pgp_fingerprints(void)
 	  "one. See the manual for more information.\n"
 	  "(Note: these fingerprints have nothing to do with SSH!)\n"
 	  "\n"
-	  "PuTTY Master Key as of 2015 (RSA, 4096-bit):\n"
+	  "PuTTY Master Key as of " PGP_MASTER_KEY_YEAR
+          " (" PGP_MASTER_KEY_DETAILS "):\n"
 	  "  " PGP_MASTER_KEY_FP "\n\n"
-	  "Original PuTTY Master Key (RSA, 1024-bit):\n"
-	  "  " PGP_RSA_MASTER_KEY_FP "\n"
-	  "Original PuTTY Master Key (DSA, 1024-bit):\n"
-	  "  " PGP_DSA_MASTER_KEY_FP "\n", stdout);
+	  "Previous Master Key (" PGP_PREV_MASTER_KEY_YEAR
+          ", " PGP_PREV_MASTER_KEY_DETAILS "):\n"
+	  "  " PGP_PREV_MASTER_KEY_FP "\n", stdout);
 }
 
-void console_provide_logctx(void *logctx)
+static void console_logging_error(LogPolicy *lp, const char *string)
 {
-    console_logctx = logctx;
+    /* Ordinary Event Log entries are displayed in the same way as
+     * logging errors, but only in verbose mode */
+    fprintf(stderr, "%s\n", string);
+    fflush(stderr);
 }
 
-void logevent(void *frontend, const char *string)
+static void console_eventlog(LogPolicy *lp, const char *string)
 {
-    log_eventlog(console_logctx, string);
+    /* Ordinary Event Log entries are displayed in the same way as
+     * logging errors, but only in verbose mode */
+    if (flags & FLAG_VERBOSE)
+        console_logging_error(lp, string);
 }
 
-static void console_data_untrusted(HANDLE hout, const char *data, int len)
+StripCtrlChars *console_stripctrl_new(
+    Seat *seat, BinarySink *bs_out, SeatInteractionContext sic)
+{
+    return stripctrl_new(bs_out, false, 0);
+}
+
+static void console_write(HANDLE hout, ptrlen data)
 {
     DWORD dummy;
-    /* FIXME: control-character filtering */
-    WriteFile(hout, data, len, &dummy, NULL);
+    WriteFile(hout, data.ptr, data.len, &dummy, NULL);
 }
 
-int console_get_userpass_input(prompts_t *p,
-                               const unsigned char *in, int inlen)
+int console_get_userpass_input(prompts_t *p)
 {
-    HANDLE hin, hout;
+    HANDLE hin = INVALID_HANDLE_VALUE, hout = INVALID_HANDLE_VALUE;
     size_t curr_prompt;
 
     /*
@@ -402,23 +474,23 @@ int console_get_userpass_input(prompts_t *p,
      */
     /* We only print the `name' caption if we have to... */
     if (p->name_reqd && p->name) {
-	size_t l = strlen(p->name);
-	console_data_untrusted(hout, p->name, l);
-	if (p->name[l-1] != '\n')
-	    console_data_untrusted(hout, "\n", 1);
+	ptrlen plname = ptrlen_from_asciz(p->name);
+	console_write(hout, plname);
+        if (!ptrlen_endswith(plname, PTRLEN_LITERAL("\n"), NULL))
+	    console_write(hout, PTRLEN_LITERAL("\n"));
     }
     /* ...but we always print any `instruction'. */
     if (p->instruction) {
-	size_t l = strlen(p->instruction);
-	console_data_untrusted(hout, p->instruction, l);
-	if (p->instruction[l-1] != '\n')
-	    console_data_untrusted(hout, "\n", 1);
+	ptrlen plinst = ptrlen_from_asciz(p->instruction);
+	console_write(hout, plinst);
+        if (!ptrlen_endswith(plinst, PTRLEN_LITERAL("\n"), NULL))
+	    console_write(hout, PTRLEN_LITERAL("\n"));
     }
 
     for (curr_prompt = 0; curr_prompt < p->n_prompts; curr_prompt++) {
 
 	DWORD savemode, newmode;
-        int len;
+        size_t len;
 	prompt_t *pr = p->prompts[curr_prompt];
 
 	GetConsoleMode(hin, &savemode);
@@ -429,20 +501,17 @@ int console_get_userpass_input(prompts_t *p,
 	    newmode |= ENABLE_ECHO_INPUT;
 	SetConsoleMode(hin, newmode);
 
-	console_data_untrusted(hout, pr->prompt, strlen(pr->prompt));
+	console_write(hout, ptrlen_from_asciz(pr->prompt));
 
         len = 0;
         while (1) {
             DWORD ret = 0;
-            BOOL r;
 
             prompt_ensure_result_size(pr, len * 5 / 4 + 512);
 
-            r = ReadFile(hin, pr->result + len, pr->resultsize - len - 1,
-                         &ret, NULL);
-
-            if (!r || ret == 0) {
-                len = -1;
+            if (!ReadFile(hin, pr->result + len, pr->resultsize - len - 1,
+                          &ret, NULL) || ret == 0) {
+                len = (size_t)-1;
                 break;
             }
             len += ret;
@@ -456,12 +525,10 @@ int console_get_userpass_input(prompts_t *p,
 
 	SetConsoleMode(hin, savemode);
 
-	if (!pr->echo) {
-	    DWORD dummy;
-	    WriteFile(hout, "\r\n", 2, &dummy, NULL);
-	}
+	if (!pr->echo)
+            console_write(hout, PTRLEN_LITERAL("\r\n"));
 
-        if (len < 0) {
+        if (len == (size_t)-1) {
             return 0;                  /* failure due to read error */
         }
 
@@ -471,10 +538,9 @@ int console_get_userpass_input(prompts_t *p,
     return 1; /* success */
 }
 
-void frontend_keypress(void *handle)
-{
-    /*
-     * This is nothing but a stub, in console code.
-     */
-    return;
-}
+static const LogPolicyVtable default_logpolicy_vt = {
+    console_eventlog,
+    console_askappend,
+    console_logging_error,
+};
+LogPolicy default_logpolicy[1] = {{ &default_logpolicy_vt }};
