@@ -9,9 +9,9 @@
 FontSpec *platform_default_fontspec(const char *name)
 {
     if (!strcmp(name, "Font"))
-        return fontspec_new("Courier New", 0, 10, ANSI_CHARSET);
+        return fontspec_new("Courier New", false, 10, ANSI_CHARSET);
     else
-        return fontspec_new("", 0, 0, 0);
+        return fontspec_new("", false, 0, 0);
 }
 
 Filename *platform_default_filename(const char *name)
@@ -29,6 +29,11 @@ char *platform_default_s(const char *name)
     return NULL;
 }
 
+bool platform_default_b(const char *name, bool def)
+{
+    return def;
+}
+
 int platform_default_i(const char *name, int def)
 {
     return def;
