@@ -34,17 +34,17 @@ struct filereq_tag {
  * save==1 -> GetSaveFileName; save==0 -> GetOpenFileName
  * `state' is optional.
  */
-BOOL request_file(filereq *state, OPENFILENAME *of, int preserve, int save)
+bool request_file(filereq *state, OPENFILENAME *of, bool preserve, bool save)
 {
     TCHAR cwd[MAX_PATH]; /* process CWD */
-    BOOL ret;
+    bool ret;
 
     /* Get process CWD */
     if (preserve) {
 	DWORD r = GetCurrentDirectory(lenof(cwd), cwd);
 	if (r == 0 || r >= lenof(cwd))
 	    /* Didn't work, oh well. Stop trying to be clever. */
-	    preserve = 0;
+	    preserve = false;
     }
 
     /* Open the file requester, maybe setting lpstrInitialDir */
@@ -142,12 +142,12 @@ void pgp_fingerprints(void)
 		"one. See the manual for more information.\n"
 		"(Note: these fingerprints have nothing to do with SSH!)\n"
 		"\n"
-                "PuTTY Master Key as of 2015 (RSA, 4096-bit):\n"
+                "PuTTY Master Key as of " PGP_MASTER_KEY_YEAR
+                " (" PGP_MASTER_KEY_DETAILS "):\n"
                 "  " PGP_MASTER_KEY_FP "\n\n"
-                "Original PuTTY Master Key (RSA, 1024-bit):\n"
-                "  " PGP_RSA_MASTER_KEY_FP "\n"
-                "Original PuTTY Master Key (DSA, 1024-bit):\n"
-                "  " PGP_DSA_MASTER_KEY_FP,
+                "Previous Master Key (" PGP_PREV_MASTER_KEY_YEAR
+                ", " PGP_PREV_MASTER_KEY_DETAILS "):\n"
+                "  " PGP_PREV_MASTER_KEY_FP,
 		"PGP fingerprints", MB_ICONINFORMATION | MB_OK,
 		HELPCTXID(pgp_fingerprints));
 }
@@ -160,11 +160,10 @@ void pgp_fingerprints(void)
 char *GetDlgItemText_alloc(HWND hwnd, int id)
 {
     char *ret = NULL;
-    int size = 0;
+    size_t size = 0;
 
     do {
-	size = size * 4 / 3 + 512;
-	ret = sresize(ret, size, char);
+        sgrowarray_nm(ret, size, size);
 	GetDlgItemText(hwnd, id, ret, size);
     } while (!memchr(ret, '\0', size-1));
 
@@ -321,7 +320,7 @@ void split_into_argv(char *cmdline, int *argc, char ***argv,
     p = cmdline; q = outputline; outputargc = 0;
 
     while (*p) {
-	int quote;
+	bool quote;
 
 	/* Skip whitespace searching for start of argument. */
 	while (*p && isspace(*p)) p++;
@@ -331,7 +330,7 @@ void split_into_argv(char *cmdline, int *argc, char ***argv,
 	outputargv[outputargc] = q;
 	outputargstart[outputargc] = p;
 	outputargc++;
-	quote = 0;
+	quote = false;
 
 	/* Copy data into the argument until it's finished. */
 	while (*p) {
