@@ -16,4 +16,4 @@ void platform_get_x11_auth(struct X11Display *disp, Conf *conf)
 	x11_get_auth_from_authfile(disp, xauthpath);
 }
 
-const int platform_uses_x11_unix_by_default = FALSE;
+const bool platform_uses_x11_unix_by_default = false;
